import os
import time
import pdb
import tracemalloc
import numpy as np
from termcolor import colored
from typing import Callable, Tuple, Set, Dict
from types import SimpleNamespace
import argparse
from loguru import logger
import torch
import torch.nn.functional as F


def setup_minimal_example(config: Dict):
    config["CONFIG_PATH"] = __file__
    config["NOW"] = ""


class PyTorchPhaseManager:
    """
    A context manager for managing the training and evaluation phases of a PyTorch model.

    Args:
    - phase (str): The phase to manage ('train' or 'val').

    Methods:
    - __init__(self, phase)
    - __enter__(self)
    - __exit__(self, *args)
    - training_preprocess(self)
    - training_postprocess(self)
    - evaluation_preprocess(self)
    - evaluation_postprocess(self)

    Required variables:
    - model
    - optimizer
    - loss

    Remark:
    - assumption for the loss function is that it takes arguments in order of:
      (model, ..., config).
    """

    def __init__(
        self,
        model: torch.nn.Module,
        optimizer: torch.optim,
        loss_func: Callable,
        loss_args: Tuple,
        phase: str,
        config: dict,
    ):
        """
        Initializes a PyTorchPhaseManager object with the specified phase.

        Args:
        - phase (str): The phase to manage ('train' or 'val').
        """

        self.config = config
        self.max_norm = config["MAX_NORM"]
        self.model = model
        self.optimizer = optimizer
        self.phase = phase
        self.loss = None
        self.loss_func = loss_func
        self.loss_args = loss_args
        self.scaler = torch.cuda.amp.GradScaler()

    def __enter__(self):
        """
        Enters the context for the current phase and performs any necessary preprocessing.

        Returns:
        - self: The PyTorchPhaseManager object.
        """

        if "train" in self.phase:
            self.mode = torch.enable_grad()
            self.training_preprocess()
        elif "val" in self.phase:
            self.mode = torch.no_grad()
            self.evaluation_preprocess()
        else:
            raise ValueError(f"Invalid phase: {self.phase}")

        self.mode.__enter__()

        return self

    def __exit__(self, *args):
        """
        Exits the context for the current phase and performs any necessary postprocessing.

        Args:
        - *args: The arguments passed to __exit__.
        """
        if "train" in self.phase:
            self.training_postprocess()
            self.grad_norm = self.monitor_gradient()
        elif "val" in self.phase:
            self.evaluation_postprocess()
            self.grad_norm = torch.tensor(0.0)
        else:
            raise ValueError(f"Invalid phase: {self.phase}")

        self.mode.__exit__(*args)

    def training_preprocess(self):
        """
        Sets the model to training mode and clears the gradients of the optimizer.
        """

        self.model.train()

        self.optimizer.zero_grad()

    def training_postprocess(self):
        """
        Computes the gradients of the loss with respect to the model parameters using loss.backward()
        and performs the optimization step using optimizer.step().
        """
        self.scaler.scale(self.loss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()

    def evaluation_preprocess(self):
        """
        Sets the model to evaluation mode.
        """
        self.model.eval()

    def evaluation_postprocess(self):
        """
        Performs no post-processing steps, as gradients are not computed during evaluation.
        """
        pass

    def monitor_gradient(self):
        """ """
        grad_raw = [
            torch.norm(p.grad.detach())
            for p in self.model.parameters()
            if p.grad is not None
        ]

        if not grad_raw:
            grad_norm = torch.tensor(0.0)
        else:
            grad_norm = torch.norm(torch.stack(grad_raw), 2)

        return grad_norm

    def call_loss(self):
        """
        compute loss function and evaluation metrics.

        Return:
        - self.loss
        - metrics (dict)
        """
        monitored_values = self.loss_func(
            self.model, *self.loss_args, config=self.config
        )
        self.loss = monitored_values["losses"]["total_loss"]

        return monitored_values


class StopWatch:
    """
    A context manager for measureing the elapsed time of a code block.

    Usage:
        with StopWatch() as time:
            <code block>
        print('elapsed time (sec):', time.elapsed)
    """

    def __init__(self, unit="hours", label=""):
        """
        Args:
        - unit (str): specifies time unit \in {seconds, minutes, hours}
        - label (str): added to log message
        """
        if "sec" in unit:
            self.denom = 1.0
            self.unit = "seconds"
        elif "min" in unit:
            self.denom = 60.0
            self.unit = "minutes"
        elif "hour" in unit:
            self.denom = 3600.0
            self.unit = "hours"
        self.label = label

    def __enter__(self):
        self.start_time = time.time()
        tracemalloc.start()

        return self

    def __exit__(self, *args):
        _end_time = time.time()
        self.elapsed = (_end_time - self.start_time) / self.denom

        _end_memory, _end_memory_peak = tracemalloc.get_traced_memory()
        tracemalloc.stop()
        self.memory_usage = _end_memory / (1024**2)  # convert to megabytes

        logger.info(f"elapsed time: {self.elapsed: 6f} {self.unit}. ({self.label})")
        logger.info(f"Memory usage: {self.memory_usage:.2f} megabytes")


class ErrorHandler:
    """
    A context manager that catches specified exceptions and invokes the pdb debugger.
    Can be useful for debugging purpose.

    Attributes:
    exception (Exception): The type of exception to catch. Defaults to Exception.

    Example:
        with ErrorHandler():
            some_computation_to_monitor()

    """

    def __init__(self, exception=Exception):
        self.exception = exception

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Invokes pdb.set_trace() if the caught exception matches the specified exception type.

        Args:
        exc_type (type): The type of the caught exception.
        exc_value (Exception): The caught exception.
        traceback (traceback): The traceback object for the caught exception.

        Returns:
        bool: True if the exception was caught and handled, False otherwise.
        """
        if exc_type is not None and issubclass(exc_type, self.exception):
            pdb.set_trace()
            return True
        return False


def convert_torch_to_numpy(data):
    """
    Recursively convert all values in a hierarchical python dictionary, a PyTorch tensor, a numpy array,
    or a list of dictionaries, tensors, or numpy arrays to numpy arrays. If a value is a PyTorch tensor
    on cpu or cuda device, it will be converted to a numpy array.

    Args:
        data (dict, tensor, array, or list): A hierarchical python dictionary, a PyTorch tensor,
        a numpy array, or a list of dictionaries, tensors, or numpy arrays whose values will be converted to numpy arrays.

    Returns:
        dict, array, list: A copy of the input dictionary, tensor, array or list with all values converted to numpy arrays.
    """
    if isinstance(data, dict):
        result = {}
        for key, value in data.items():
            result[key] = convert_torch_to_numpy(value)
        return result
    elif isinstance(data, torch.Tensor):
        # handle data in traini or eval phase
        data = data.detach() if data.requires_grad else data

        # handle data on GPU or CPU
        if data.is_cuda:
            return data.cpu().numpy()
        else:
            return data.numpy()
    elif isinstance(data, list):
        result = []
        for item in data:
            result.append(convert_torch_to_numpy(item))
        return result
    else:  # numpy array or else
        return data


def create_directories_and_log(dict_path):
    """
    A simple helper function create directories and output log message.
    """
    if not os.path.exists(dict_path):
        os.makedirs(dict_path)
        logger.info(
            colored("A new log directory was created: ", "yellow") + f"{dict_path}"
        )


def format_common_folder_structure(conf):
    """ """
    if conf.comment:
        return f"/{conf.comment}_{conf.model_backbone}_{conf.now}"
    else:
        return f"/{conf.model_backbone}_{conf.now}"


class ConfigSubset:
    """
    A class to store the required key-value pairs as instance variables.
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key.lower(), value)


def extract_params_from_config(requirements: Set[str], config: Dict) -> SimpleNamespace:
    """
    Extract necessary (hyper)parameters for each function. The extracted
    parameters are stored as a class instance variables for easy access.
    Keys of the necessary parameters are defined in the set "requirements."
    The keys are converted to lowercase and used for instance variable names.

    Use this class to avoid accidentally overwrite the original config function -
    remember that python function gets a pointer to the original config dict,
    not a copy of it (i.e., config can be modified within a function).

    Args:
    - requirements (set): required keys for a given function.
    - config (dict): the original dictionary containing all necessary parameters.

    Returns:
    - sub_conf (ConfigSubset): required variables stored as instance variables of a class.
    """

    # assert that all the required keys exist in the config file.
    missing_keys = requirements.difference(config.keys())
    assert not missing_keys, f"Missing necessary parameters: {', '.join(missing_keys)}"

    sub_conf = SimpleNamespace(**{key.lower(): config[key] for key in requirements})
    return sub_conf


def float_formatter(variable, digits=3):
    """
    Format float variable, or string-represented float, with
    specified effective digits.

    Args:
    - variable: str, float, or else

    Returns:
    - string f'{variable:.5}' if variable is float or
      string-represented float. Return variable unchanged if else
    """

    def formatted_float(var, digits):
        # return f'{var:.%d}' % digits
        return f"{var:.{digits}}"

    if isinstance(variable, float):
        return formatted_float(variable, digits)
    elif isinstance(variable, str):
        try:
            float_variable = float(variable)
            return formatted_float(float_variable, digits)
        except:
            return variable
    else:
        return variable


def grab_gpu(config, device=None):
    # check if necessary parameters are defined in the config file
    requirements = set(["GPU"])
    conf = extract_params_from_config(requirements, config)
    # restrict visible device(s)
    os.environ["CUDA_VISIBLE_DEVICES"] = str(conf.gpu)
    # Get cpu or gpu device for training.
    if not device:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    print("Using {} device".format(device))
    return torch.device(device)


def make_directories(path):
    if not os.path.exists(path):
        logger.info("Path '{}' does not exist.".format(path))
        logger.warning("Make directory: " + path)
        os.makedirs(path)


def fix_random_seed(config):
    # check if necessary parameters are defined in the config file
    requirements = set(["IS_SEED", "SEED"])
    conf = extract_params_from_config(requirements, config)

    if conf.is_seed:
        np.random.seed(conf.seed)
        torch.manual_seed(conf.seed)
        logger.info(
            "Numpy and Pytorch's random seeds are fixed: seed=" + str(conf.seed)
        )

    else:
        logger.info("Random seed is not fixed.")


def set_cpu_workers(config: dict) -> None:
    """
    Set the number of CPU threads (workers).
    This number is set equal to the number of num_workers in PyTorch DataLoader.
    """
    # check if necessary parameters are defined in the config file
    requirements = set(["NUM_WORKERS"])
    conf = extract_params_from_config(requirements, config)
    torch.set_num_threads(np.max([1, conf.num_workers]))


def compile_comment(config):
    """
    Name log file(s) based on selected parameters.
    """
    requirements = set(
        [
            "PRUNER_NAME",
            "ORDER_SPRT",
            "LLLR_VERSION",
            "PARAM_LLR_LOSS",
            "PARAM_MULTIPLET_LOSS",
            "IS_ADAPTIVE_LOSS",
            "EXP_PHASE",
            "SUBPROJECT_NAME_PREFIX",
            "SUBPROJECT_NAME_SUFFIX",
            "COMMENT",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    str_pruner = f"_{conf.pruner_name}pruner"
    str_main = (
        f"order{conf.order_sprt}_lllr{conf.lllr_version}{float_formatter(conf.param_llr_loss)}"
        f"_mult{float_formatter(conf.param_multiplet_loss)}_adap{conf.is_adaptive_loss}"
    )
    name = f"{conf.comment}_" + str_main
    if "tuning" in conf.exp_phase.lower():
        name += str_pruner
    return name


def compile_subproject_name(config):
    """
    define the subproject name, which is used to create a folder for training logs.
    """
    assert type(config) == dict
    requirements = set(
        [
            "DATA_SEPARATION",
            "OPTIMIZATION_TARGET",
            "SUBPROJECT_NAME_PREFIX",
            "SUBPROJECT_NAME_SUFFIX",
            "EXP_PHASE",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    str_main = (
        conf.subproject_name_prefix
        + f'offset{conf.data_separation.replace(".", "_")}_optim{conf.optimization_target}'
        + conf.subproject_name_suffix
        + "_"
        + conf.exp_phase
    )

    return str_main


def compile_directory_paths(config):
    common_path = f"{config['LOG_PATH']}{config['SUBPROJECT_NAME']}/"
    config["ROOT_TBLOGS"] = common_path + f"{config['TB_DIRNAME']}"
    config["ROOT_DBLOGS"] = common_path + f"{config['DB_DIRNAME']}"
    config["ROOT_CKPTLOGS"] = common_path + f"{config['CKPT_DIRNAME']}"
    config["ROOT_STDOUTLOGS"] = common_path + f"{config['STDOUT_DIRNAME']}"
    config["PATH_RESUME"] = (
        common_path + f"{config['CKPT_DIRNAME']}/{config['SUBPROJECT_TO_RESUME']}"
    )


def parse_args(config: dict) -> None:
    """
    Parse the command-line arguments for the training script.

    Args:
    - config (dict): a dictionary containing default hyperparameters.

    Return:
    - None: no need to return because this function receives a pointer to config dict.
    """

    def fix_dependent_parameters(config: dict) -> None:
        """
        Some config parameters are dependent on other parameters.
        Fix them based on the updated config file.
        """
        config["NUM_EPOCHS"] = (
            config["NUM_ITER"] * config["BATCH_SIZE"] // config["NUM_TRAIN"]
        )

    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "GPU",
            "NUM_ITER",
            "EXP_PHASE",
            "OPTIMIZATION_TARGET",
            "EXP_PHASE",
            "MODEL_BACKBONE",
            "SUBPROJECT_NAME_PREFIX",
            "DATA_SEPARATION",
            "NUM_TRIALS",
            "IS_LOAD_ONTO_MEMORY",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    parser = argparse.ArgumentParser(
        description="Sequential Density Ratio Estimation by the SPRT-TANDEM "
    )
    parser.add_argument("-g", "--gpu", type=int, default=conf.gpu, help="#gpu")
    parser.add_argument(
        "-d",
        "--data_separation",
        type=str,
        default=conf.data_separation,
        help="separation of Gaussian dataset.",
    )
    parser.add_argument(
        "-t", "--num_trials", type=int, default=conf.num_trials, help="#trials"
    )
    parser.add_argument(
        "-i", "--num_iter", type=int, default=conf.num_iter, help="#iterations"
    )
    parser.add_argument(
        "-e",
        "--exp_phase",
        type=str,
        default=conf.exp_phase,
        help='phase of an experiment, "try," "tuning," or "stat"',
    )
    parser.add_argument(
        "-m",
        "--model",
        type=str,
        default=conf.model_backbone,
        help='model backbone, "LSTM", or "Transformer"',
    )
    parser.add_argument(
        "-o",
        "--optimize",
        type=str,
        default=conf.optimization_target,
        help='optimization target: "MABS", "MacRec", "ausat_confmx", or "ALL"',
    )
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        default=conf.subproject_name_prefix,
        help="subproject name",
    )
    parser.add_argument(
        "--flip_memory_loading",
        action="store_true",
        help="set a boolean flag indicating whether to load onto memory",
    )

    args = parser.parse_args()

    # overwrite config dict with the provided command-line arguments
    config["GPU"] = args.gpu
    config["DATA_SEPARATION"] = args.data_separation
    config["NUM_TRIALS"] = args.num_trials
    config["NUM_ITER"] = args.num_iter
    config["EXP_PHASE"] = args.exp_phase
    config["MODEL_BACKBONE"] = args.model
    config["OPTIMIZATION_TARGET"] = args.optimize
    config["SUBPROJECT_NAME_PREFIX"] = args.name
    config["SUBPROJECT_NAME"] = compile_subproject_name(config)
    config["IS_LOAD_ONTO_MEMORY"] = (
        not config["IS_LOAD_ONTO_MEMORY"]
        if args.flip_memory_loading
        else config["IS_LOAD_ONTO_MEMORY"]
    )
    fix_dependent_parameters(config)
    compile_directory_paths(config)


# Functions for graph plotting
def restrict_classes(llrs, labels, list_classes):
    """
    Args:
        list_classes: A list of integers.
    Remark:
        (batch, duration, num classes, num classes)
        -> (< batch, duration, num classes, num classes)
    """
    if list_classes == []:
        return llrs, labels

    assert torch.min(labels).item() <= min(list_classes)
    assert max(list_classes) <= torch.max(labels).item()

    ls_idx = []
    for itr_cls in list_classes:
        ls_idx.append(torch.squeeze(torch.where(labels == itr_cls)))
    idx = torch.cat(ls_idx, dim=0)
    idx, _ = torch.sort(idx)

    llrs_rest = torch.gather(
        llrs, 0, idx.unsqueeze(1).expand(-1, llrs.shape[1], -1, -1)
    )
    lbls_rest = torch.gather(
        labels, 0, idx.unsqueeze(1).expand(-1, labels.shape[1], -1, -1)
    )

    return llrs_rest, lbls_rest


def extract_positive_row(llrs, labels):
    """Extract y_i-th rows of LLR matrices.
    Args:
        llrs: (batch, duraiton, num classes, num classes)
        labels: (batch,)
    Returns:
        llrs_posrow: (batch, duration, num classes)
    """
    llrs_shape = llrs.shape
    duration = llrs_shape[1]
    num_classes = llrs_shape[2]

    labels_oh = F.one_hot(labels, num_classes=num_classes)
    # (batch, num cls)
    labels_oh = labels_oh.reshape(-1, 1, num_classes, 1)
    labels_oh = labels_oh.repeat(1, duration, 1, 1)
    # (batch, duration, num cls, 1)
    llrs_pos = llrs * labels_oh
    # (batch, duration, num cls, num cls)
    llrs_posrow = torch.sum(llrs_pos, dim=2)
    # (batch, duration, num cls): = LLR_{:, :, y_i, :}

    return llrs_posrow


def add_max_to_diag(llrs):
    """
    Args:
        llrs: (batch, duration, num classes, num classes)
    Returns:
        llrs_maxdiag: (batch, duration, num classes, num classes),
            max(|llrs|) is added to diag of llrs.
    """
    num_classes = llrs.shape[2]

    llrs_abs = torch.abs(llrs)
    llrs_max = torch.max(llrs_abs)
    # max |LLRs|
    tmp = torch.eye(num_classes) * llrs_max
    tmp = tmp.unsqueeze(0).unsqueeze(0)
    tmp = tmp.to(llrs.device)
    llrs_maxdiag = llrs + tmp

    return llrs_maxdiag



from typing import Any, Dict, Tuple

import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data as data
from loguru import logger
from optuna.trial import Trial
from termcolor import colored
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from ..models.losses import compute_loss_and_metrics
from ..models.optimizers import initialize_optimizer
from ..models.temporal_integrators import import_model, load_pretrained_states
from ..utils.checkpoint import update_and_save_result
from ..utils.data_processing import lmdb_dataloaders, move_data_to_device
from ..utils.hyperparameter_tuning import report_to_pruner
from ..utils.logging import (
    get_tb_writer,
    log_training_results,
    save_config_info,
    save_sdre_imgs,
)
from ..utils.misc import (
    PyTorchPhaseManager,
    convert_torch_to_numpy,
    fix_random_seed,
    set_cpu_workers,
)
from ..utils.performance_metrics import accumulate_performance, training_setup


@logger.catch
def prepare_for_training(
    config: Dict[str, Any], device: str
) -> Tuple[nn.Module, optim.Optimizer, Dict[str, data.DataLoader], SummaryWriter]:
    """
    Prepare the network and data for training.

    Args:
    - trial: An object used by Optuna to generate trial parameters. Unused if not using Optuna.
    - config: A dictionary containing various configuration settings.
    - device: The device to run the computation on, e.g. 'cpu' or 'cuda'.

    Returns:
    - model: The initialized network.
    - optimizer: The optimizer used to update the network parameters during training.
    - data_loaders: A dictionary containing the train, validation, and test data loaders.
    - tb_writer: A SummaryWriter object for logging training progress to TensorBoard.
    """
    # accurate matrix multiplication when working with 32-bit floating-point values
    torch.set_float32_matmul_precision("high")

    # optimize cuda computation if network structure is static, sacrificing reproducibility
    torch.backends.cudnn.benchmark = True

    # set number of CPU threads (workers)
    set_cpu_workers(config)

    # set random seeds (optional)
    fix_random_seed(config)

    # save config.py and config (dict) for reproducibility
    save_config_info(config)

    # setup Tensorboard writer
    tb_writer = get_tb_writer(config)

    # initialize the network or load a pretrained one
    # tb_writer is optional: provide if you want to have a model graph on TB
    model = import_model(config, device, tb_writer=None)

    # setup the optimizer
    model, optimizer = initialize_optimizer(model, config)

    # (optional) load pretrained state_dicts
    load_pretrained_states(model, optimizer, config)

    # load train, val, and test data
    data_loaders = lmdb_dataloaders(config)

    return model, optimizer, data_loaders, tb_writer


@logger.catch
def iterating_over_dataset(
    trial: Trial,
    model: nn.Module,
    optimizer: optim.Optimizer,
    device: str,
    data_loaders: Dict[str, data.DataLoader],
    tb_writer: SummaryWriter,
    config: Dict[str, Any],
    best: Tuple[float, ...],
    global_step: int,
    phase: str,
) -> Tuple[Dict[str, float], int, Tuple[float], Tuple[float, ...]]:
    """
    Runs the model for a given phase (training, validation, or test) and logs results.

    Args:
    - trial: An object used by Optuna to generate trial parameters. Unused if not using Optuna.
    - model: The PyTorch model to be run.
    - optimizer: The PyTorch optimizer used to update model parameters during training.
    - device: The device to run the computation on, e.g. 'cpu' or 'cuda'.
    - data_loaders: A dictionary containing train, validation, and test data loaders.
    - tb_writer: A TensorBoard SummaryWriter to log results.
    - config: A dictionary containing model configuration parameters.
    - best: A tuple of the current best performance metrics.
    - global_step: The current iteration number.
    - phase: The phase for which the model is being run. Can be 'train', 'validation', or 'test'.

    Returns:
    - performance_metrics: A dictionary containing the performance metrics.
    - global_step: The updated iteration number.
    - last_example_to_plot: An object for plotting results.
    - best: A tuple of the updated best performance metrics.
    """

    is_train, iter_num, performance_metrics, barcolor = training_setup(phase, config)

    for local_step, data in tqdm(
        enumerate(data_loaders[phase]),
        mininterval=2,
        total=iter_num,
        desc=colored(f"{phase} in progress...", barcolor),
        colour=barcolor,
        leave=False,
    ):
        x_batch, y_batch, gt_llrs_batch = move_data_to_device(data, device)  # type: ignore

        # a periodic validation is inserted between training iterations.
        # This function is recursively called under is_train=True, with phase='eval'.
        if is_train and global_step % config["VALIDATION_STEP"] == 0:
            best, global_step = run_one_epoch(
                trial,
                model,
                optimizer,
                device,
                data_loaders,
                tb_writer,
                config,
                best,
                global_step=global_step,
                phase="val",
            )

        # Train phase: run preprocesses (model.train(), optimizer.zero_grad()) and postprocesses (optimizer.step(), loss.backward())
        # Eval phase: run preprocesses (model.eval()), enter torch.no_grad() mode, no postprocess
        with PyTorchPhaseManager(
            model,
            optimizer,
            phase=phase,
            loss_func=compute_loss_and_metrics,
            loss_args=(x_batch, y_batch),
            config=config,
        ) as p:
            monitored_values = p.call_loss()
            p.loss = monitored_values["losses"]["total_loss"]

        # Store performance metrics
        performance_metrics = accumulate_performance(
            performance_metrics,
            y_batch,
            gt_llrs_batch,
            monitored_values,
            phase_manager=p,
        )

        # log results periodically. Return immediately when the condition is not met
        # performance_metrics is reset at training loop to avoid running out of memory
        performance_metrics = log_training_results(
            tb_writer,
            model,
            local_step,
            global_step,
            iter_num,
            config,
            performance_metrics,
            phase,
        )

        # for figures
        if local_step == iter_num - 1 and "val" in phase:
            last_example_to_plot = convert_torch_to_numpy(
                [
                    y_batch,
                    gt_llrs_batch,
                    monitored_values["llrs"],
                    monitored_values["thresholds"],
                ]
            )
        else:
            last_example_to_plot = None

        # increment global step that traces training process. skip if eval phase
        global_step = global_step + 1 if "train" in phase else global_step

    return performance_metrics, global_step, last_example_to_plot, best


@logger.catch
def eval_postprocess(
    model: nn.Module,
    optimizer: optim.Optimizer,
    best: Tuple[float, ...],
    performance_metrics: Dict[str, float],
    last_example_to_plot: Tuple[float],
    config: Dict[str, Any],
    tb_writer: SummaryWriter,
    global_step: int,
) -> Tuple[float, ...]:
    """
    Updates the status and saves a checkpoint.

    Args:
    - model: The PyTorch model.
    - optimizer: The PyTorch optimizer.
    - best: A tuple of the current best performance metrics.
    - performance_metrics: A dictionary containing the performance metrics.
    - last_example_to_plot: An object for plotting results.
    - config: A dictionary containing model configuration parameters.
    - global_step: The current iteration number.

    Returns:
    - best: A tuple of the updated best performance metrics.
    """
    # update status and save a checkpoint
    best = update_and_save_result(
        model, optimizer, config, best, performance_metrics, global_step=global_step
    )

    # save trajectory figures if needed
    save_sdre_imgs(
        config, best, tb_writer, global_step, last_example_to_plot, performance_metrics
    )

    return best


def run_one_epoch(
    trial: Trial,
    model: nn.Module,
    optimizer: optim.Optimizer,
    device: str,
    data_loaders: Dict[str, data.DataLoader],
    tb_writer: SummaryWriter,
    config: Dict[str, Any],
    best: Tuple[float, ...],
    global_step: int,
    phase: str = "train",
) -> Tuple[Tuple[float, ...], int]:
    """
    Runs one epoch of the model.

    Args:
    - trial: An optuna Trial object.
    - model: The PyTorch model.
    - optimizer: The PyTorch optimizer.
    - device: The device on which to run the computation.
    - data_loaders: A dictionary containing the data loaders for the train, validation, and test sets.
    - tb_writer: A tensorboard SummaryWriter to log results.
    - config: A dictionary containing model configuration parameters.
    - best: A tuple of the current best performance metrics.
    - global_step: The current iteration number.
    - phase: The phase for which the model is being run. Can be 'train', 'validation', or 'test'.

    Returns:
    - best: A tuple of the updated best performance metrics.
    - global_step: The current iteration number.
    """
    (
        performance_metrics,
        global_step,
        last_example_to_plot,
        best,
    ) = iterating_over_dataset(
        trial,
        model,
        optimizer,
        device,
        data_loaders,
        tb_writer,
        config,
        best,
        global_step,
        phase,
    )

    # postprocesses at an eval phase
    if "val" in phase:
        best = eval_postprocess(
            model,
            optimizer,
            best,
            performance_metrics,
            last_example_to_plot,
            config,
            tb_writer,
            global_step,
        )

        # optuna pruner for early stopping
        report_to_pruner(trial, best, global_step, config)

    return best, global_step



# Assuption: this file is placed inside a directory named "config", under the project root directory.

from ..utils.misc import compile_subproject_name, compile_directory_paths

########################## USER EDITABLE BLOCK STARTS ######################################

# Base info
CONFIG_PATH = __file__
LOG_PATH = CONFIG_PATH[: CONFIG_PATH.find("config")] + "logs/"

# Data info
NUM_CLASSES = 2
DATA_SEPARATION = "2.0"
FEAT_DIM = 2
DATA_PATH = "../data/"
DATA_FOLDER = (
    f"SequentialGaussian_{NUM_CLASSES}class_offset{DATA_SEPARATION}_dim{FEAT_DIM}"
)
NUM_TRAIN = 40000  # 300000  # 25000 # 19000
NUM_VAL = 4000  # 30000  # 4000 # # 990
NUM_TEST = 4000  # 30000  # 1000 # 10
TRAIN_DATA = f"{DATA_PATH}{DATA_FOLDER}/train_{NUM_TRAIN}"
VAL_DATA = f"{DATA_PATH}{DATA_FOLDER}/val_{NUM_VAL}"
TEST_DATA = f"{DATA_PATH}{DATA_FOLDER}/test_{NUM_TEST}"
DATA_NAMES = ("data", "label", "llr")
IS_SHUFFLE = True  # whether to shuffle training data. It may cause significant overhead
TB_DIRNAME = "TensorBoard_events"
DB_DIRNAME = "Optuna_databases"
CKPT_DIRNAME = "checkpoints"
STDOUT_DIRNAME = "stdout_logs"

SUBPROJECT_NAME_PREFIX = "public_"
SUBPROJECT_NAME_SUFFIX = ""
COMMENT = ""  # for the log file name
OPTIMIZATION_CANDIDATES = (
    "macrec",
    "mabs",
    "ausat_confmx",
)
OPTIMIZATION_TARGET = "all"  # "ausat_confmx"
EXP_PHASE = "try"  # 'try', 'tuning', or 'stat'

# if EXP_PHASE=='tuning', these params will be overwritten
# with a new parameter sampled by Optuna.
MODEL_BACKBONE = "LSTM"
# non-negative value and must be less than time_steps of the data.
ORDER_SPRT = 5
PARAM_MULTIPLET_LOSS = 0.6
LLLR_VERSION = "LSEL"  # LLLR or LSEL
PARAM_LLR_LOSS = 0.3
IS_ADAPTIVE_LOSS = False

# Reproduce a particular trial or the best trial from specified subproject.
REPRODUCE_TRIAL = False  # 'best', trial index (int) or False
# Load pretrained weight from .pt file specified below.
IS_RESUME = False
SUBPROJECT_TO_RESUME = "_20230330_214519224/ckpt_step3500_target_ausat_confmx0.0011.pt"

# of frequent use
GPU = 0
BATCH_SIZE = 100
NUM_TRIALS = 1000
NUM_EPOCHS = 10
NUM_ITER = NUM_EPOCHS * NUM_TRAIN // BATCH_SIZE  # e.g., 20 * 25000 // 100 = 5000
TRAIN_DISPLAY_STEP = 250
VALIDATION_STEP = 250
# hyperband, median, percentile, etc... set to 'none' if no pruner is needed.
PRUNER_NAME = "percentile"
PRUNER_STARTUP_TRIALS = 100
PRUNER_WARMUP_STEPS = 2000
PRUNER_INTERVAL_STEPS = 500

############################## USER EDITABLE BLOCK ENDS #######################################

info_for_subproject_name = {
    "DATA_SEPARATION": DATA_SEPARATION,
    "OPTIMIZATION_TARGET": OPTIMIZATION_TARGET,
    "MODEL_BACKBONE": MODEL_BACKBONE,
    "PRUNER_NAME": PRUNER_NAME,
    "ORDER_SPRT": ORDER_SPRT,
    "LLLR_VERSION": LLLR_VERSION,
    "PARAM_LLR_LOSS": PARAM_LLR_LOSS,
    "PARAM_MULTIPLET_LOSS": PARAM_MULTIPLET_LOSS,
    "IS_ADAPTIVE_LOSS": IS_ADAPTIVE_LOSS,
    "EXP_PHASE": EXP_PHASE,
    "SUBPROJECT_NAME_PREFIX": SUBPROJECT_NAME_PREFIX,
    "SUBPROJECT_NAME_SUFFIX": SUBPROJECT_NAME_SUFFIX,
}

SUBPROJECT_NAME = compile_subproject_name(info_for_subproject_name)

subconfig = {
    "LOG_PATH": LOG_PATH,
    "TB_DIRNAME": TB_DIRNAME,
    "DB_DIRNAME": DB_DIRNAME,
    "CKPT_DIRNAME": CKPT_DIRNAME,
    "STDOUT_DIRNAME": STDOUT_DIRNAME,
    "SUBPROJECT_NAME": SUBPROJECT_NAME,
    "SUBPROJECT_TO_RESUME": SUBPROJECT_TO_RESUME,
}

compile_directory_paths(subconfig)

config = {
    "VERBOSE": True,
    "CONFIG_PATH": CONFIG_PATH,
    "IS_SEED": False,
    "SEED": 7,
    "GPU": GPU,
    "MAX_NORM": 50000,  # used for gradient clipping
    # for logging
    "NAME_DATASET": "Multivariate_Gaussian",
    "COMMENT": COMMENT,
    "IS_SAVE_FIGURE": True,
    "DATA_SEPARATION": DATA_SEPARATION,
    "SUBPROJECT_NAME_PREFIX": SUBPROJECT_NAME_PREFIX,
    "SUBPROJECT_NAME_SUFFIX": SUBPROJECT_NAME_SUFFIX,
    # SDRE datasets
    "TRAIN_DATA": TRAIN_DATA,
    "VAL_DATA": VAL_DATA,
    "TEST_DATA": TEST_DATA,
    "DATA_NAMES": DATA_NAMES,
    "IS_SHUFFLE": IS_SHUFFLE,
    "NUM_TRAIN": NUM_TRAIN,
    "NUM_VAL": NUM_VAL,
    "NUM_TEST": NUM_TEST,
    "NUM_CLASSES": NUM_CLASSES,
    "FEAT_DIM": FEAT_DIM,
    # SPRT-TANDEM parameters
    "ORDER_SPRT": ORDER_SPRT,
    "TIME_STEPS": 50,
    "LLLR_VERSION": LLLR_VERSION,  # LSEL
    "OBLIVIOUS": False,  # Miyagawa & Ebihara, ICML2021
    "PARAM_MULTIPLET_LOSS": PARAM_MULTIPLET_LOSS,
    "PARAM_LLR_LOSS": PARAM_LLR_LOSS,
    # Network specs
    # LSTM (B2Bsqrt-TANDEM) or Transformer (TANDEMformer)
    "MODEL_BACKBONE": MODEL_BACKBONE,
    "ACTIVATION_OUTPUT": "B2Bsqrt",
    "ACTIVATION_FC": "relu",
    "ALPHA": 1.0,  # for B2Bsqrt activation function
    "IS_NORMALIZE": False,  # whether to use Normalization or not
    "IS_POSITIONAL_ENCODING": True,
    "IS_TRAINABLE_ENCODING": True,
    # For LSTM backbone
    "ACTIVATION_INPUT": "sigmoid",
    "WIDTH_LSTM": 64,
    # For TRANSFORMER backbone
    "NUM_BLOCKS": 2,  # transformer block
    "NUM_HEADS": 2,
    "DROPOUT": 0.1,
    "FF_DIM": 64,
    "MLP_UNITS": 64,
    # for performance
    "IS_COMPILE": False,  # torch.compile (pytorch > 2.0)
    "MODE": "reduce-overhead",  # 'reduce-overhead', 'default', or 'max-autotune'
    "NUM_WORKERS": 2,  # num_workers argument for pytorch dataloader
    # whether to load dataset onto memory at initialization.
    "IS_LOAD_ONTO_MEMORY": True,
    # SAT curve
    "NUM_THRESH": 1000,
    # "linspace", "logspace", "unirandom", or "lograndom".
    "SPARSITY": "logspace",
    "BETA": 1.0,
    "IS_ADAPTIVE_LOSS": IS_ADAPTIVE_LOSS,
    # Training parameters
    "OPTIMIZATION_TARGET": OPTIMIZATION_TARGET,
    "OPTIMIZATION_CANDIDATES": OPTIMIZATION_CANDIDATES,
    "PRUNER_STARTUP_TRIALS": PRUNER_STARTUP_TRIALS,
    "PRUNER_WARMUP_STEPS": PRUNER_WARMUP_STEPS,
    "PRUNER_INTERVAL_STEPS": PRUNER_INTERVAL_STEPS,
    # hyperband, median, percentile, etc... set to 'none' if no pruner is needed.
    "PRUNER_NAME": PRUNER_NAME,
    "MAX_TO_KEEP": 1,
    "EXP_PHASE": EXP_PHASE,
    "REPRODUCE_TRIAL": REPRODUCE_TRIAL,
    "NUM_TRIALS": NUM_TRIALS,
    "BATCH_SIZE": BATCH_SIZE,
    "LEARNING_RATE": 1e-4,
    "LR_DECAY_STEPS": [
        100000000,
    ],
    "WEIGHT_DECAY": 0.0,
    "OPTIMIZER": "adam",
    "NUM_EPOCHS": NUM_EPOCHS,
    "NUM_ITER": NUM_ITER,  # 7500
    "TRAIN_DISPLAY_STEP": TRAIN_DISPLAY_STEP,
    "VALIDATION_STEP": VALIDATION_STEP,
    "IS_RESUME": IS_RESUME,
    ####### Optuna hyperparameter search space #######
    # Key values starting "SPACE_" containing parameter search space information.
    # PARAM_SPACE: "float", "int", or "categorical".
    # - float: use suggest_float to suggest a float of range [LOW, HIGH], separated by STEP.
    #   if LOG=True, a float is sampled from logspace but you shall set STEP=None.
    # - int: use suggest_int to suggest an int of range [LOW, HIGH], separated by STEP.
    #   STEP should be divisor of the range, otherwise HIGH will be automatically modified.
    #   if LOG=True, an int is sampled from logspace but you shall set STEP=None.
    # - categorical: use suggest_categorical to select one category from CATEGORY_SET.
    #   Note that if the parameter is continuous (e.g., 1, 2, 3, ..., or 1.0, 0.1, 0.001, ...),
    #   it is adviseable to use float or int space because suggest_categorical treats
    #   each category independently.
    "SPACE_MODEL_BACKBONE": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": ["LSTM", "Transformer"],  # 'LSTM', 'Transformer'
    },
    # common to all backbones
    "SPACE_WEIGHT_DECAY": {
        "PARAM_SPACE": "float",
        "LOW": 0.0,
        "HIGH": 0.001,
        "STEP": 0.0001,
        "LOG": False,  # log is preferable but it doesn't allow LOW==0.0
    },
    "SPACE_LEARNING_RATE": {
        "PARAM_SPACE": "float",
        "LOW": 0.000001,
        "HIGH": 0.001,
        "STEP": None,
        "LOG": True,
    },
    "SPACE_LLLR_VERSION": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": ["LLLR", "LSEL"],
    },
    "SPACE_ACTIVATION_FC": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": ["B2Bsqrt", "B2BsqrtV2", "B2Bcbrt", "tanh", "relu", "gelu"],
    },
    "SPACE_PARAM_MULTIPLET_LOSS": {
        "PARAM_SPACE": "float",
        "LOW": 0.0,
        "HIGH": 1.0,
        "STEP": 0.1,
        "LOG": False,
    },
    "SPACE_PARAM_LLR_LOSS": {
        "PARAM_SPACE": "float",
        "LOW": 0.0,
        "HIGH": 1.0,
        "STEP": 0.1,
        "LOG": False,
    },
    "SPACE_IS_ADAPTIVE_LOSS": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": [True, False],
    },
    "SPACE_ORDER_SPRT": {
        "PARAM_SPACE": "int",
        "LOW": 0,
        "HIGH": 5,  # 10
        "STEP": 1,
        "LOG": False,
    },
    "SPACE_IS_POSITIONAL_ENCODING": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": [True, False],
    },
    "SPACE_IS_TRAINABLE_ENCODING": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": [True, False],
    },
    "SPACE_OPTIMIZER": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": ["rmsprop", "adam"],
    },
    "SPACE_IS_NORMALIZE": {"PARAM_SPACE": "categorical", "CATEGORY_SET": [True, False]},
    "SPACE_NUM_THRESH": {
        "PARAM_SPACE": "int",
        "LOW": 500,
        "HIGH": 1500,
        "STEP": 500,
        "LOG": False,
    },
    "SPACE_SPARSITY": {
        "PARAM_SPACE": "categorical",
        "CATEGORY_SET": ["linspace", "logspace", "unirandom", "lograndom"],
    },
    # hyperparams specific to each backbone
    # B2Bsqrt-TANDEM
    "LSTM": {
        "SPACE_ACTIVATION_OUTPUT": {
            "PARAM_SPACE": "categorical",
            "CATEGORY_SET": ["B2Bsqrt", "B2BsqrtV2", "B2Bcbrt", "tanh", "sigmoid"],
        },
        "SPACE_ACTIVATION_INPUT": {
            "PARAM_SPACE": "categorical",
            "CATEGORY_SET": ["tanh", "sigmoid"],
        },
        "SPACE_WIDTH_LSTM": {
            "PARAM_SPACE": "int",
            "LOW": 16,
            "HIGH": 128,
            "STEP": 1,
            "LOG": True,
        },
    },
    # TANDEMformer
    "TRANSFORMER": {
        "SPACE_NUM_BLOCKS": {
            "PARAM_SPACE": "int",
            "LOW": 1,
            "HIGH": 2,
            "STEP": 1,
            "LOG": False,
        },
        "SPACE_NUM_HEADS": {  # num_heads need to be a divisor of feat_dim (=embed_dim)
            "PARAM_SPACE": "int",
            "LOW": 2,
            "HIGH": 4,
            "STEP": 2,
            "LOG": False,
        },
        "SPACE_DROPOUT": {
            "PARAM_SPACE": "float",
            "LOW": 0.0,
            "HIGH": 0.5,
            "STEP": 0.1,
            "LOG": False,
        },
        "SPACE_FF_DIM": {
            "PARAM_SPACE": "int",
            "LOW": 32,
            "HIGH": 64,
            "STEP": 32,
            "LOG": False,
        },
        "SPACE_MLP_UNITS": {
            "PARAM_SPACE": "int",
            "LOW": 32,
            "HIGH": 64,
            "STEP": 32,
            "LOG": False,
        },
    },
    # Some combination of hyperparameters are incompatible or ineffective.
    "FORBIDDEN_PARAM_SETS": {
        "loss_all_zero": {
            "SPACE_PARAM_MULTIPLET_LOSS": 0.0,
            "SPACE_PARAM_LLR_LOSS": 0.0,
        },
    },
}

config.update(subconfig)



import shutil
import datetime
import pdb
import numpy as np
import colorsys
from torch.utils.tensorboard import SummaryWriter
import matplotlib.pyplot as plt
import matplotlib
from tqdm import tqdm
from loguru import logger
from ..utils.performance_metrics import (
    initialize_performance_metrics,
    summarize_performance,
)
from ..utils.misc import (
    extract_params_from_config,
    create_directories_and_log,
    format_common_folder_structure,
    compile_comment,
    convert_torch_to_numpy,
)


def save_config_info(config):
    """
    Save config.py and config (dict) for reproducibility.
    """
    requirements = set(["CONFIG_PATH", "DIR_CONFIGS", "NOW", "EXP_PHASE"])
    conf = extract_params_from_config(requirements, config)

    comment = compile_comment(config)

    shutil.copyfile(conf.config_path, conf.dir_configs + f"/config_{conf.now}_orig.py")
    with open(conf.dir_configs + f"/config_{conf.now}_cooked.py", "w") as f:
        f.write(
            "# This is the actual config dict used for the training.\n"
            "# Simply import this file for reproducing the results.\n\n"
        )
        f.write("config = {\n")
        for key, value in config.items():
            if isinstance(value, str):
                f.write(f'  "{key}": "{value}", \n')
            else:
                f.write(f'  "{key}": {value}, \n')
        f.write("}\n")
    with open(conf.dir_configs + f"/{comment}.txt", "w") as f:
        f.write("")


def create_log_folders(config):
    """
    Create log folders for earch trial:
    - checkpoint log folder
    - tensorboard event log folder
    - stdout log folder
    Note that an optuna database log folder is created for each subproject, not each trial.
    For definition of optuna database log folder, see utils.hyperparameter_tuning:run_optuna.
    """

    def setup_checkpoint_log_folder(config, conf):
        config[
            "DIR_CKPTLOGS"
        ] = f"{conf.root_ckptlogs}" + format_common_folder_structure(conf)
        config["DIR_CONFIGS"] = config["DIR_CKPTLOGS"] + "/configs"
        create_directories_and_log(config["DIR_CONFIGS"])

    def setup_tensorboard_log_folder(config, conf):
        config["DIR_TBLOGS"] = f"{conf.root_tblogs}" + format_common_folder_structure(
            conf
        )
        create_directories_and_log(config["DIR_TBLOGS"])

    def setup_stdout_log_folder(config, conf):
        # do not make individual folder because one trial contains one .log file anyways
        # + format_common_folder_structure(conf)[:-1]
        config["DIR_STDOUTLOGS"] = f"{conf.root_stdoutlogs}"
        config[
            "STDOUTLOG_NAME"
        ] = f"{config['DIR_STDOUTLOGS']}/{format_common_folder_structure(conf)}.log"
        create_directories_and_log(config["DIR_STDOUTLOGS"])

    # get a timestamp for each trial
    config["NOW"] = datetime.datetime.now().strftime("%Y%m%d_%H%M%S%f")[:-3]

    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "ROOT_CKPTLOGS",
            "ROOT_TBLOGS",
            "ROOT_STDOUTLOGS",
            "EXP_PHASE",
            "COMMENT",
            "CONFIG_PATH",
            "NOW",
            "MODEL_BACKBONE",
            "ORDER_SPRT",
            "LLLR_VERSION",
            "PARAM_LLR_LOSS",
            "PARAM_MULTIPLET_LOSS",
            "IS_ADAPTIVE_LOSS",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    setup_checkpoint_log_folder(config, conf)
    setup_tensorboard_log_folder(config, conf)
    setup_stdout_log_folder(config, conf)


def setup_logger(config):
    def tqdm_compatible_logger(
        logger, logpath, rotation="10 MB", enqueue=True, level="DEBUG"
    ):
        """
        loguru logger outputs to console and log file specified with logpath.
        lambda function is used to wrap tqdm.write(), which is required to
        output message onto the console without disrupting tqdm progress bar.

        Args:
        - logger: loguru logger object

        Return:
        - None
        """
        logger.remove()
        logger.add(
            lambda msg: tqdm.write(msg, end=""),
            enqueue=enqueue,
            level=level,
            colorize=True,
            backtrace=True,
            diagnose=True,
        )
        logger.add(logpath, rotation=rotation, enqueue=enqueue, level=level)

    requirements = set(["STDOUTLOG_NAME"])
    conf = extract_params_from_config(requirements, config)

    tqdm_compatible_logger(logger, conf.stdoutlog_name)


def investigate_log(config, is_stop_at_error):
    """
    Investigate .log file and count WARNING and ERROR sign.
    Raise ValueError when non-zero ERROR are found.
    """
    requirements = set(["STDOUTLOG_NAME"])
    conf = extract_params_from_config(requirements, config)

    with open(conf.stdoutlog_name, "r") as f:
        contents = f.read()
    num_warning = contents.count("WARNING")
    num_error = contents.count("ERROR")
    logger.info(f"{num_warning=}/{num_error=} in the .log file.")

    if num_error > 0:
        if is_stop_at_error:
            raise ValueError("Found ERROR! Check the .log file for debug!")
        else:
            logger.warning("Found ERROR! check the .log file for debug!")
    elif num_warning > 0:
        logger.warning("Found WARNING! check the .log file for debug!")


class ContexualLogger:
    """
    A context manager that sets up a logger before executing a code block, and
    investigates the logger after the code block has completed.

    Usage:
    config = initialize_setups()
    with LoggingContextManager(config):
        training_my_neural_network(config)
    """

    def __init__(self, config, is_stop_at_error: bool = True):
        self.config = config
        self.is_stop_at_error = is_stop_at_error

    def __enter__(self):
        """
        Sets up a logger before entering the with block.

        Returns:
        self: The current LoggingContextManager instance.
        """
        setup_logger(self.config)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Investigates the logger after exiting the with block.

        Args:
        exc_type: The type of exception raised (if any).
        exc_value: The value of the exception raised (if any).
        traceback: The traceback object associated with the exception (if any).
        """
        investigate_log(self.config, self.is_stop_at_error)


def get_tb_writer(config):
    """
    Get a PyTorch-builtin TensorBoard writer.

    Args:
    - config (dict): a dictionary containing key "DIR_TBLOGS" and
      its value specifying the path to tensorboard log file.

    Return:
    - SummaryWriter(): PyTorch SummaryWriter object for logging.
    """
    # check if necessary parameters are defined in the config file
    requirements = set(["DIR_TBLOGS"])
    conf = extract_params_from_config(requirements, config)

    return SummaryWriter(log_dir=conf.dir_tblogs)


def log_training_results(
    tb_writer,
    model,
    local_step: int,
    global_step: int,
    data_num: int,
    config: dict,
    performance_metrics: dict,
    phase: str,
):
    """
    Log training and evaluation results at:
    - the end of validation phase
    - evary config['TRAIN_DISPLAY_STEP'] training iteration
    """

    def subsample_dict(performance_metrics, selected_keys):
        return {k: v for k, v in performance_metrics.items() if k in selected_keys}

    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "VERBOSE",
            "IS_ADAPTIVE_LOSS",
            "PARAM_MULTIPLET_LOSS",
            "PARAM_LLR_LOSS",
            "WEIGHT_DECAY",
            "TRAIN_DISPLAY_STEP",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    # log train results only periodically
    if "train" in phase.lower() and global_step % conf.train_display_step != 0:
        # reset the performance metrics to save GPU memory
        return initialize_performance_metrics()

    if "val" in phase.lower() and local_step != data_num - 1:
        return performance_metrics

    _model = model

    # summarize performance_metrics
    performance_metrics = summarize_performance(performance_metrics)

    # select values of interest (I might like to modify here later)
    subset_metrics = subsample_dict(
        performance_metrics,
        [
            "losses",
            "mean_abs_error",
            "mean_macro_recall",
            "ausat_from_confmx",
            "grad_norm",
        ],
    )

    # output to console and log
    if conf.verbose:
        logger.info("")
        logger.info("Global Step ={:7d}".format(global_step))
        logger.info(
            "{} mean(MacRec)_t: {}".format(
                phase, performance_metrics["mean_macro_recall"]
            )
        )
        logger.info(
            "{} mean ABSerr: {}".format(phase, performance_metrics["mean_abs_error"])
        )
        logger.info(
            "{} ausat from confmx: {}".format(
                phase, performance_metrics["ausat_from_confmx"]
            )
        )
        if conf.is_adaptive_loss:
            adaptive_loss_weights = {
                "MCE": _model.adaptive_loss_weights[0],
                "LLLR": _model.adaptive_loss_weights[1],
                "weight_decay": _model.adaptive_loss_weights[2],
            }
            logger.info(
                "{} MCE loss:{:7.5f} * {} * {:7.5f}".format(
                    phase,
                    performance_metrics["losses"]["multiplet_crossentropy (MCE)"],
                    str(conf.param_multiplet_loss),
                    adaptive_loss_weights["MCE"],
                )
            )
            logger.info(
                "{} LLLR :{:7.5f} * {} * {:7.5f}".format(
                    phase,
                    performance_metrics["losses"]["LLR_estimation (LLRE)"],
                    str(conf.param_llr_loss),
                    adaptive_loss_weights["LLLR"],
                )
            )
            logger.info(
                "{} weight decay:{:7.5f} * {} * {:7.5f}".format(
                    phase,
                    performance_metrics["losses"]["weight_decay"],
                    str(conf.weight_decay),
                    adaptive_loss_weights["weight_decay"],
                )
            )
        else:
            logger.info(
                "{} MCE loss:{:7.5f} * {}".format(
                    phase,
                    performance_metrics["losses"]["multiplet_crossentropy (MCE)"],
                    str(conf.param_multiplet_loss),
                )
            )
            logger.info(
                "{} LLLR :{:7.5f} * {}".format(
                    phase,
                    performance_metrics["losses"]["LLR_estimation (LLRE)"],
                    str(conf.param_llr_loss),
                )
            )
            logger.info(
                "{} weight decay:{:7.5f} * {}".format(
                    phase,
                    performance_metrics["losses"]["weight_decay"],
                    str(conf.weight_decay),
                )
            )
    # write to TensorBoard
    tblog_writer(tb_writer, global_step, phase=phase, metrics=subset_metrics)

    # training phase: reset the performance metrics after the log
    # eval phase: needs postprocess, do not reset the performance metrics
    return (
        initialize_performance_metrics()
        if "train" in phase.lower()
        else performance_metrics
    )


def tblog_writer(tb_writer, global_step, phase, metrics):
    """
    Logs various values to TensorBoard.
    The keys in the kwargs dictionary will be used to create the names for the values in TensorBoard.

    Args:
    - tb_writer (torch.utils.tensorboard.SummaryWriter()): TensorBoard logger to use for logging the values.
    - global_step (int): Global step to use for logging the values.
    - phase (str): The phase of the model, used to prefix the names of the logged values.
    - kwargs (dict): Dictionary of values to log. The keys in the dictionary will be used to create the names for the values in TensorBoard.

    Returns:
    None

    Example:
    - a keyword argument loss={'LLLR': 0.2} at a training phase will be logged as:
        'training_loss/LLLR' with value 0.2.
    """

    def write_to_tensorboard(metrics):
        for key, value in metrics.items():
            if isinstance(value, dict):
                write_to_tensorboard(value)
            else:
                name = f"{phase}/{key}"
                tb_writer.add_scalar(name, value, global_step)

    assert isinstance(metrics, dict), f"metrics is not a dictionary!"
    # use a recursive function to log values
    write_to_tensorboard(metrics)


def add_prefix_to_keys(original_dict: dict, prefix: str) -> dict:
    """
    Add a prefix to all the keys in a dictionary.
    Parameters:
        original_dict (dict): The original dictionary whose keys will be modified.
        prefix (str): The prefix to add to all the keys in the original dictionary.
    Returns:
        dict: A new dictionary with the prefix added to all the keys.
    """
    assert isinstance(original_dict, dict), "original_dict should be of type dict"
    assert isinstance(prefix, str), "prefix should be of type str"
    return {prefix + key: value for key, value in original_dict.items()}


def plot_likelihood_ratio_matrix(
    llrm: np.ndarray, gt_labels: np.ndarray, num_trajectories: int = 50
) -> None:
    """
    Plot the likelihood ratio matrix for each pair of classes in llrm.

    Args:
    - llrm (ndarray): an array of shape (batch_size, num_classes, num_classes) containing the log-likelihood ratio
      matrix for each sample in the dataset.
    - gt_labels (ndarray): an array of shape (batch_size) containing the ground truth labels for the samples in the dataset.
    - num_trajectories (int, optional): the number of trajectories to plot for each pair of classes. Defaults to 50.
    """
    num_classes = llrm.shape[-1]
    colors = plt.cm.tab10(range(num_classes))

    for i in range(num_classes):
        for j in range(num_classes):
            plt.subplot(num_classes, num_classes, (j + 1) + num_classes * i)
            for k, c in zip(range(num_classes), colors):
                plt.plot(
                    np.transpose(llrm[gt_labels == k, :, i, j][:num_trajectories]),
                    color=c,
                )

    plt.subplot(num_classes, num_classes, 1)
    for (
        k,
        c,
    ) in zip(range(num_classes), colors):
        plt.plot(
            np.transpose(llrm[gt_labels == k, :, 0, 0][:1]), color=c, label=f"class {k}"
        )
    plt.legend()


def plot_llrs_with_thresholds(
    llrs: np.ndarray,
    gt_labels: np.ndarray,
    thresholds: np.ndarray,
    max_traj_per_cls_pair: int = 1,
    max_traj_per_llr: int = 1,
    threshold_level: str = "max",
    plot_labels: bool = False,
):
    """
    Args:
    - llrs (np.array): array of size (batch_size, time_steps, num_classes, num_classes)
    - gt_labels (np.array): array of size (batch_size)
    - thresholds (np.array): array of size (num_thresh, batch_size, time_steps, num_classes, num_classes)
    - max_traj_per_cls_pair (int): positive integer specifies the maximum number of trajectories per class
    - max_traj_per_llr (int): positive integer specifies the maximum number of threshold trajectories per LLR
    - threshold_level (str):
        'zero', 'smallest' or 'min': use the minimum threshold. max_traj_per_cls_pair is automatically set to 1.
        'small' use the random thresholds from the first third of the data, exept the minumum threshold.
        'medium' use the random thresholds from the second third of the data.
        'large' use the random thresholds from the last third of the data, exept the maximum threshold.
        'largest' or 'max' : use the maximum threshold. max_traj_per_cls_pair is automatically set to 1.
    """
    assert threshold_level in [
        "zero",
        "smallest",
        "min",
        "small",
        "medium",
        "large",
        "largest",
        "max",
    ]
    assert max_traj_per_cls_pair > 0
    assert thresholds.shape[1:] == llrs.shape

    num_classes = len(np.unique(gt_labels))
    num_thresh = thresholds.shape[0]

    # select thresholds of size=max_traj_per_llr
    if any(s in threshold_level for s in ["zero", "smallest", "min"]):
        dice_thresh = np.array([0])
        max_traj_per_llr = 1
    elif any(s in threshold_level for s in ["largest", "max"]):
        dice_thresh = np.array([num_thresh - 1])
        max_traj_per_llr = 1
    else:
        if "small" in threshold_level:
            start_idx = 1
            end_idx = int(num_thresh / 3)
        elif "medium" in threshold_level:
            start_idx = int(num_thresh / 3)
            end_idx = int(num_thresh * 2 / 3)
        elif "large" in threshold_level:
            start_idx = int(num_thresh * 2 / 3)
            end_idx = num_thresh - 1
        dice_thresh = np.random.choice(
            range(start_idx, end_idx), size=max_traj_per_llr, replace=False
        )

    # colors for plot
    colors = matplotlib.cm.tab10(range(num_classes))
    ex_colors = expand_color_variations(colors, max_traj_per_llr)

    # find minimum number of examples across the classes
    cls_samples = []
    for k in range(num_classes):
        cls_samples.append(len(llrs[gt_labels == k]))

    # randomly selected example data indices of size=max_traj_per_cls_pair
    dice_batch = np.random.permutation(np.min(cls_samples) - 1)[:max_traj_per_cls_pair]

    for j in range(num_classes):
        for k in range(num_classes):
            if k == j:
                continue
            for i in range(max_traj_per_llr):
                i_th = dice_thresh[i]
                label = f"class {k} vs. {j} at y={k} with thresh #{i_th}"
                color = ex_colors[k, i]
                targllr = np.transpose(llrs[gt_labels == k, :, k, j][dice_batch])
                targth = np.transpose(
                    thresholds[i_th, gt_labels == k, :, k, j][dice_batch]
                )
                plt.plot(targllr, color=color, label=label)
                plt.plot(targth, color=color)
    if plot_labels:
        handles, labels = plt.gca().get_legend_handles_labels()
        unique_label_set = dict(zip(labels, handles))
        plt.legend(unique_label_set.values(), unique_label_set.keys(), fontsize=10)


def rgb_to_hsv(color):
    return colorsys.rgb_to_hsv(color[0], color[1], color[2])


def hsv_to_rgb(color):
    return colorsys.hsv_to_rgb(color[0], color[1], color[2])


def create_color_variations(color, num_variations=5):
    hsv_color = rgb_to_hsv(color)
    variations = []
    for i in range(1, num_variations + 1):
        new_saturation = hsv_color[1] * (1 - i * 0.1)
        new_color = (hsv_color[0], new_saturation, hsv_color[2])
        variations.append(hsv_to_rgb(new_color))
    return variations


def expand_color_variations(base_colors, num_variations=5):
    num_colors = base_colors.shape[0]
    # Create the color variations
    color_matrix = np.empty((num_colors, num_variations, 4))
    for i, color in enumerate(base_colors[:, :3]):  # only take the RBG values
        variations = create_color_variations(color, num_variations)
        for j, variation in enumerate(variations):
            color_matrix[i, j, :3] = variation
            color_matrix[i, j, 3] = 1.0  # Set alpha channel to 1
    return color_matrix


def plot_example_trajectories(
    llrm, gt_labels, dice=[None, None], max_traj_per_cls_pair=10, max_example_cls=10
):
    """
    Summarize a log-likelihood ratio matrix (llrm) into one panel.

    Args:
    - llrm (float): array of size (batch_size, time_steps, num_classes, num_classes)
      log likelihood ratio matrix, either ground-truth or estimated.
    - gt_labels (int): array of size (batch_size,). gt_labels \in [0, 1, ... num_classes - 1]
      ground truth data labels.
    - dice (int or None): array of size (max_traj_per_cls_pair,) if provided.
      For selecting/reproducing trajectories to be plotted.
      - dice[0]: indices for trajectories.
      - dice[1]: indices for classes.
    - max_traj_per_cls_pair (int): positive integer specifies the maximum number of trajectories
      per class pair. e.g., if num_classes = 3, there will be 3*2=6 class pairs and thus
      6 * max_traj_per_cls_pair trajectories will be plotted in total. Note that a number of
      trajectories will be smaller than max_traj_per_cls_pair if llrm smaller number of examples.
    - max_example_cls (int): positive integer specifies the maximum number of classes to be plotted.
      classes are randomly selected if max_example_cls > num_classes.

    Returns:
    - dice (int): the dice that used to select example trajectories. Use it to plot to select
      the corresponding trajectories of different llrms (e.g., ground-truth llrm and estimated llrm).

    Example:
    plt.figure(figsize=(16.2, 10))

    plt.subplot(1,2,1)
    plt.title('Ground-truth LLR trajectories')
    dice = plot_example_trajectories(gt_llrm, y)

    plt.subplot(1,2,2)
    plt.title('Estimated LLR trajectories')
    _ = plot_example_trajectories(estimated_llrm, y, dice=dice)

    """

    # limit example
    num_classes = len(np.unique(gt_labels))
    assert np.max(gt_labels) + 1 == num_classes

    assert max_example_cls > 1, (
        "max_example_cls must be greater than 1 to ensure likelihood ratio calculation"
        " (i.e., ratio is defined with cls > 1.)"
    )

    colors = matplotlib.cm.tab10(range(num_classes))
    if all(d is None for d in dice):
        cls_samples = []
        for k in range(num_classes):
            cls_samples.append(len(llrm[gt_labels == k]))
        dice[0] = np.random.permutation(np.min(cls_samples) - 1)[:max_traj_per_cls_pair]

        # limit example classes if needed
        if num_classes > max_example_cls:
            dice[1] = np.random.permutation(num_classes)[:max_example_cls]
        else:
            dice[1] = range(num_classes)
    else:
        assert (
            len(dice[0]) == max_traj_per_cls_pair
        ), "length of dice[0] must be equal to max_traj_per_cls_pair!"

    for j in dice[1]:
        for k, c in zip(dice[1], colors):
            if k == j:
                continue
            plt.plot(
                np.transpose(llrm[gt_labels == k, :, k, j][dice[0]]),
                color=c,
                label=f"class {k} vs. {j} at y={k}",
            )
    # labels
    handles, labels = plt.gca().get_legend_handles_labels()
    unique_label_set = dict(zip(labels, handles))
    plt.legend(unique_label_set.values(), unique_label_set.keys(), fontsize=10)

    return dice


def save_sdre_imgs(
    config, target, tb_writer, global_step, last_example_to_plot, performance_metrics
):
    """
    Save the results of sequential density ratio estimation (SDRE).
    """
    # check if necessary parameters are defined in the config file
    requirements = set(
        [
            "IS_SAVE_FIGURE",
            "OPTIMIZATION_TARGET",
            "PARAM_MULTIPLET_LOSS",
            "PARAM_LLR_LOSS",
            "NUM_CLASSES",
            "BATCH_SIZE",
        ]
    )
    conf = extract_params_from_config(requirements, config)

    if conf.is_save_figure == False:
        return

    y_batch, gt_llrs, llrs, thresholds = last_example_to_plot

    # not necessarily == conf.batch_size if batch_size is not a divisor of data number
    try:
        assert len(y_batch) == len(gt_llrs) == len(llrs)
    except:
        logger.info(
            "Catch the error might caused by class_num in y_batch is less than conf.num_classes"
        )
        pdb.set_trace()

    performance_metrics = convert_torch_to_numpy(performance_metrics)
    aesthetic_num_traj = np.max([1, 20 // conf.num_classes])

    plt.rcParams["font.size"] = 15
    fig, ax = plt.subplots(figsize=(20, 6))

    plt.subplot(1, 3, 1)
    plt.title("Estimated LLR trajectories")
    dice = plot_example_trajectories(
        llrs, y_batch, max_traj_per_cls_pair=aesthetic_num_traj
    )
    plt.subplot(1, 3, 2)
    plt.title("Ground-truth LLR trajectories")
    if not gt_llrs is None:
        _ = plot_example_trajectories(
            gt_llrs, y_batch, dice, max_traj_per_cls_pair=aesthetic_num_traj
        )
    else:
        # plt.text('ground-truth LLR is not available')
        pass

    plt.subplot(1, 3, 3)
    plt.scatter(performance_metrics["hitting_time"], performance_metrics["sns_conf"])
    plt.title("SAT curve from the confusion matrix")

    targval = str(np.array([t for t in target]))
    name = f"_mla{conf.param_multiplet_loss}"
    f"{conf.param_llr_loss}_target{conf.optimization_target}{targval}"

    # savie in the TensorBoard log
    tb_writer.add_figure(name, fig, global_step=global_step)

    plt.close("all")
    logger.info("Figures saved.")



import warnings

import numpy as np
import torch
from torchmetrics import ConfusionMatrix

from ..utils.misc import (
    ErrorHandler,
    convert_torch_to_numpy,
    extract_params_from_config,
)


def confusion_matrix(preds, labels, num_classes):
    """
    Calculate confusion matrix.
    """
    task = (
        "binary" if num_classes == 2 else "multiclass"
    )  # ConfusionMatrix also supports multilabel
    confmat = ConfusionMatrix(task=task, num_classes=num_classes)
    return confmat(preds, labels)


# Functions: logit to confusion matrix
def logits_to_confmx(logits, labels):
    """Calculate the confusion matrix from logits.
    Args:
        logits: A logit Tensor with shape (batch, num classes).
        labels: A non-one-hot label Tensor with shape (batch,).
    Returns:
        confmx: A Tensor with shape (num classes, num classes).
    """
    logits_shape = logits.shape  # (batch, num classes)
    num_classes = logits_shape[-1]

    # First order_sprt+1 frames
    preds = torch.argmax(logits, dim=-1, keepdim=False).to(torch.int32)
    confmx = confusion_matrix(preds, labels, num_classes)

    return confmx


# Functions related to differentials of mean hitting times
######################################################
def multiply_diff_mht(acc_eta, hittimes):
    """Multiplies acc_eta by Delta t (ver x hor = area).
    # Args
    acc_eta: ...
    hittingtimes: Shape=(num thresh, batch)
    time_steps: An int.
    """
    time_steps = hittimes.shape[0]
    assert time_steps == acc_eta.shape[0], "Length of time_steps mismatched!"

    _device = acc_eta.device
    # Calc differentials of mean hitting times
    mht = torch.mean(hittimes, dim=1) if len(hittimes.shape) > 1 else hittimes
    # from (num thresh, batch)
    # to (num thresh,)
    _mht = torch.cat(
        [mht[1:], torch.tensor([time_steps], dtype=torch.float32, device=_device)],
        dim=0,
    )
    diff_mht = _mht - mht

    aucsat = torch.sum(diff_mht * torch.squeeze(acc_eta))
    aucsat_norm = aucsat / time_steps  # normalize

    return aucsat_norm


def initialize_performance_metrics():
    """ """
    return {
        "losses": [],
        "mean_abs_error": [],
        "sns_conf": [],
        "seqconfmx_llr": 0,
        "hitting_time": [],
        "ausat_from_confmx": None,
        "grad_norm": [],
    }


def training_setup(phase: str, config):
    """
    Returns:
    - is_train (bool): train flag to decide if trainable weights are updated.
    - iter_num (int): an index of the last full-size batch.
    - performance_metrics (dict): dictionary of model performance metrics
                                  initialized with zeros or empty list.
    - barcolor (str): specifies tqdm bar color for each phase.
    """
    # check if necessary parameters are defined in the config file
    requirements = set(["NUM_TRAIN", "NUM_VAL", "NUM_TEST", "BATCH_SIZE"])
    conf = extract_params_from_config(requirements, config)

    # suppress warnings at gradient clipping
    warnings.filterwarnings(
        "ignore",
        message="The .grad attribute of a Tensor that is not a leaf Tensor is being accessed",
    )

    is_train = True if "train" in phase.lower() else False
    barcolor = "cyan" if "train" in phase.lower() else "yellow"

    if "train" in phase.lower():
        iter_num = np.ceil(conf.num_train / conf.batch_size).astype("int")
    elif "val" in phase.lower():
        iter_num = np.ceil(conf.num_val / conf.batch_size).astype("int")
    elif "test" in phase.lower():
        iter_num = np.ceil(conf.num_test / conf.batch_size).astype("int")
    else:
        raise ValueError("Unknown phase!")
    performance_metrics = initialize_performance_metrics()

    return is_train, iter_num, performance_metrics, barcolor


def accumulate_performance(
    performance_metrics, y_batch, gt_llrs_batch, monitored_values, phase_manager=None
):
    """ """

    performance_metrics["losses"].append(monitored_values["losses"])
    performance_metrics["mean_abs_error"].append(
        calc_llr_abserr(monitored_values["llrs"], gt_llrs_batch)
    )  # (batch_size, time_steps)
    performance_metrics["seqconfmx_llr"] += llr_sequential_confmx(
        monitored_values["llrs"], y_batch
    )
    performance_metrics["hitting_time"].append(monitored_values["mht"])
    performance_metrics["sns_conf"].append(monitored_values["sns_from_confmx"])
    if phase_manager:
        performance_metrics["grad_norm"].append(phase_manager.grad_norm)

    return performance_metrics


def summarize_performance(performance_metrics):
    """ """

    def calc_macrec(confmx):
        """
        Args:
        - confmx: sequential confusion matrix
        Return:
        - macrec: macro-averaged recall
        """
        return torch.mean(seqconfmx_to_macro_ave_sns(confmx))

    # average
    performance_metrics["losses"] = average_dicts(performance_metrics["losses"])
    performance_metrics["mean_macro_recall"] = calc_macrec(
        performance_metrics["seqconfmx_llr"]
    )
    performance_metrics["mean_abs_error"] = torch.mean(
        torch.cat(performance_metrics["mean_abs_error"], dim=0)
    )
    performance_metrics["hitting_time"] = torch.mean(
        torch.stack(performance_metrics["hitting_time"]), dim=0
    )  # each entry has a size of (500)
    performance_metrics["sns_conf"] = torch.mean(
        torch.stack(performance_metrics["sns_conf"]), dim=0
    )  # each entry has a size of (500, 1)
    performance_metrics["ausat_from_confmx"] = multiply_diff_mht(
        performance_metrics["sns_conf"], performance_metrics["hitting_time"]
    )
    performance_metrics["grad_norm"] = torch.mean(
        torch.stack(performance_metrics["grad_norm"])
    )
    # to numpy
    performance_metrics = convert_torch_to_numpy(performance_metrics)

    return performance_metrics


def multiplet_sequential_confmx(logits_concat, labels_concat):
    """Calculate the confusion matrix for each frame from logits. Lite.
    Args:
        logits_concat: A logit Tensor with shape
            (batch, (time_steps - order_sprt), order_sprt + 1, num_classes).
            This is the output from
            utils.data_processing.sequential_concat(logit_slice, y_slice).
        labels_concat: A non-one-hot label Tensor with shape (batch,).
            This is the output from
            utils.data_processing.sequential_conclogit_slice, y_slice).
    Return:
        seqconfmx_mult: A Tensor with shape
        (time_steps, num classes, num classes). This is the series of
        confusion matrices computed from multiplets.
    Remark:
        e.g., order_sprt = 5, time_steps = 20:
            confusion matrix for frame001 is given by the 001let of frame001
            confusion matrix for frame002 is given by the 002let of frame002
            ...
            confusion matrix for frame005 is given by the 004let of frame004
            confusion matrix for frame005 is given by the 005let of frame005
            confusion matrix for frame006 is given by the 005let of frame006 computed from frame002-006
            confusion matrix for frame007 is given by the 005let of frame007 computed from frame003-007
            ...
            confusion matrix for frame019 is given by the 005let of frame019 computed from frame015-019
            confusion matrix for frame020 is given by the 005let of frame020 computed from frame016-020
    """
    logits_concat_shape = (
        logits_concat.shape
    )  # (batch, (time_steps - order_sprt), order_sprt + 1, num classes)
    num_classes = logits_concat_shape[-1]

    # First order_sprt+1 frames
    # (batch, order_sprt + 1, num classes)
    logits_concat_former = logits_concat[:, 0, :, :]

    for iter_idx in range(logits_concat_shape[2]):
        preds_former = torch.argmax(logits_concat_former[:, iter_idx, :], dim=-1).to(
            torch.int32
        )  # (batch,)
        if iter_idx == 0:
            seqconfmx_mult = confusion_matrix(
                preds_former, labels_concat, num_classes
            ).to(torch.int32)
            seqconfmx_mult = torch.unsqueeze(seqconfmx_mult, 0)
        else:
            seqconfmx_mult = torch.cat(
                [
                    seqconfmx_mult,
                    torch.unsqueeze(
                        confusion_matrix(preds_former, labels_concat, num_classes).to(
                            torch.int32
                        ),
                        0,
                    ),
                ],
                dim=0,
            )

    # Latter time_steps-order_sprt-1 frames
    # (batch, time_steps-order_sprt-1, num classes)
    logits_concat_latter = logits_concat[:, 1:, -1, :]

    for iter_idx in range(logits_concat_shape[1] - 1):
        preds_latter = torch.argmax(logits_concat_latter[:, iter_idx, :], dim=-1).to(
            torch.int32
        )  # (batch,)
        seqconfmx_mult = torch.cat(
            [
                seqconfmx_mult,
                torch.unsqueeze(
                    confusion_matrix(preds_latter, labels_concat, num_classes).to(
                        torch.int32
                    ),
                    dim=0,
                ),
            ],
            dim=0,
        )

    return seqconfmx_mult


# Functions: confusion matrix to metric
def confmx_to_metrics(confmx):
    """Calc confusion-matrix-based performance metrics.
    Args:
        confmx: A confusion matrix Tensor
            with shape (num classes, num classes).
    Return:
        dict_metrics: A dictionary of dictionaries of performance metrics.
            E.g., sensitivity of class 3 is dics_metrics["SNS"][3],
            and accuracy is dict_metrics["ACC"]["original"]
    Remark:
        - SNS: sensitivity, recall, true positive rate
        - SPC: specificity, true negative rate
        - PRC: precision
        - ACC: accuracy
        - BAC: balanced accuracy
        - F1: F1 score
        - GM: geometric mean
        - MCC: Matthews correlation coefficient. May cause overflow.
        - MK: markedness
        - e.g., The classwise accuracy of class i is dict_metric["SNS"][i].
        - "dict_metrics" contains some redundant metrics;
          e.g., for binary classification,
          dict_metric["SNS"]["macro"] = dict_metric["BAC"][0]
          = dict_metric["BAC"][1] = ...
        - Macro-averaged metrics are more robust to class-imbalance
          than micro-averaged ones, but note that macro-averaged metrics
          are sometimes equal to be ACC.
        - Most of the micro-averaged metrics are equal to or equivalent to ACC.
    """
    confmx = confmx.to(torch.int64)  # prevent from overflowing
    num_classes = confmx.shape[0]
    dict_metrics = {
        "SNS": dict(),
        "SPC": dict(),
        "PRC": dict(),
        "ACC": dict(),
        "BAC": dict(),
        "F1": dict(),
        "GM": dict(),
        "MCC": dict(),
        "MK": dict(),
    }
    TP_tot = 0
    TN_tot = 0
    FP_tot = 0
    FN_tot = 0

    # Calc 2x2 confusion matrices out of the multiclass confusion matrix
    for i in range(num_classes):
        # Initialization
        TP = 0
        TN = 0
        FP = 0
        FN = 0

        # Calc TP, TN, FP, FN for class i
        TP = confmx[i, i]
        for j in range(num_classes):
            if j == i:
                continue
            FP += confmx[j, i]
            FN += confmx[i, j]
            for k in range(num_classes):
                if k == i:
                    continue
                TN += confmx[j, k]

        # Calc performance metrics of class i
        dict_metrics["SNS"][i] = TP / (TP + FN) if TP + FN != 0 else 0.0
        dict_metrics["SPC"][i] = TN / (TN + FP) if TN + FP != 0 else 0.0
        dict_metrics["PRC"][i] = TP / (TP + FP) if TP + FP != 0 else 0.0
        dict_metrics["ACC"][i] = (
            (TP + TN) / (TP + FN + TN + FP) if TP + FN + TN + FP != 0 else 0.0
        )
        dict_metrics["BAC"][i] = (dict_metrics["SNS"][i] + dict_metrics["SPC"][i]) / 2
        dict_metrics["F1"][i] = (
            2
            * (dict_metrics["PRC"][i] * dict_metrics["SNS"][i])
            / (dict_metrics["PRC"][i] + dict_metrics["SNS"][i])
            if dict_metrics["PRC"][i] + dict_metrics["SNS"][i] != 0
            else 0.0
        )
        dict_metrics["GM"][i] = np.sqrt(dict_metrics["SNS"][i] * dict_metrics["SPC"][i])
        dict_metrics["MCC"][i] = (
            ((TP * TN) - (FP * FN))
            / (np.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)))
            if (TP + FP) * (TP + FN) * (TN + FP) * (TN + FN) != 0
            else 0.0
        )
        dict_metrics["MK"][i] = (
            dict_metrics["PRC"][i] + (TN / (TN + FN)) - 1
            if TN + FN != 0
            else dict_metrics["PRC"][i] - 1
        )

        TP_tot += TP
        TN_tot += TN
        FP_tot += FP
        FN_tot += FN

    # Calc micro- and macro- averaged metrics
    # sometimes returns nan. please fix it
    dict_metrics["SNS"]["macro"] = np.mean(
        [dict_metrics["SNS"][i] for i in range(num_classes)]
    )
    dict_metrics["SNS"]["micro"] = (
        TP_tot / (TP_tot + FN_tot) if TP_tot + FN_tot != 0 else 0.0
    )  # = original ACC.
    dict_metrics["SPC"]["macro"] = np.mean(
        [dict_metrics["SPC"][i] for i in range(num_classes)]
    )
    dict_metrics["SPC"]["micro"] = (
        TN_tot / (TN_tot + FP_tot) if TN_tot + FP_tot != 0 else 0.0
    )
    dict_metrics["PRC"]["macro"] = np.mean(
        [dict_metrics["PRC"][i] for i in range(num_classes)]
    )
    dict_metrics["PRC"]["micro"] = (
        TP_tot / (TP_tot + FP_tot) if TP_tot + FP_tot != 0 else 0.0
    )  # = original ACC.
    dict_metrics["ACC"]["macro"] = np.mean(
        [dict_metrics["ACC"][i] for i in range(num_classes)]
    )
    dict_metrics["ACC"]["micro"] = (
        (TP_tot + TN_tot) / (TP_tot + FN_tot + TN_tot + FP_tot)
        if TP_tot + FN_tot + TN_tot + FP_tot != 0
        else 0.0
    )
    dict_metrics["ACC"]["original"] = (
        (num_classes / 2) * dict_metrics["ACC"]["micro"]
    ) - ((num_classes - 2) / 2)
    dict_metrics["BAC"]["macro"] = np.mean(
        [dict_metrics["BAC"][i] for i in range(num_classes)]
    )
    dict_metrics["BAC"]["micro"] = (
        dict_metrics["SNS"]["micro"] + dict_metrics["SPC"]["micro"]
    ) / 2
    dict_metrics["F1"]["macro"] = np.mean(
        [dict_metrics["F1"][i] for i in range(num_classes)]
    )
    dict_metrics["F1"]["micro"] = (
        2
        * (dict_metrics["PRC"]["micro"] * dict_metrics["SNS"]["micro"])
        / (dict_metrics["PRC"]["micro"] + dict_metrics["SNS"]["micro"])
        if dict_metrics["PRC"]["micro"] + dict_metrics["SNS"]["micro"] != 0
        else 0.0
    )  # = original ACC.
    dict_metrics["GM"]["macro"] = np.mean(
        [dict_metrics["GM"][i] for i in range(num_classes)]
    )
    dict_metrics["GM"]["micro"] = np.sqrt(
        dict_metrics["SNS"]["micro"] * dict_metrics["SPC"]["micro"]
    )
    dict_metrics["MCC"]["macro"] = np.mean(
        [dict_metrics["MCC"][i] for i in range(num_classes)]
    )
    dict_metrics["MCC"]["micro"] = (
        ((TP_tot * TN_tot) - (FP_tot * FN_tot))
        / (
            np.sqrt(
                (TP_tot + FP_tot)
                * (TP_tot + FN_tot)
                * (TN_tot + FP_tot)
                * (TN_tot + FN_tot)
            )
        )
        if (TP_tot + FP_tot) * (TP_tot + FN_tot) * (TN_tot + FP_tot) * (TN_tot + FN_tot)
        != 0
        else 0.0
    )
    dict_metrics["MK"]["macro"] = np.mean(
        [dict_metrics["MK"][i] for i in range(num_classes)]
    )
    dict_metrics["MK"]["micro"] = (
        dict_metrics["PRC"]["micro"] + (TN_tot / (TN_tot + FN_tot)) - 1
        if TN_tot + FN_tot != 0
        else 0.0
    )

    return dict_metrics


def average_dicts(dict_list):
    """
    This function takes a list of dictionaries as input and returns a new dictionary that
    contains the average of the values for each key across all the input dictionaries.
    The keys in the input dictionaries should be identical.

    Args:
    - dict_list(list of dictionary)

    Returns:
    - averaged dict_list

    Raise:
    - AssertionError if not all dictionaries have the same keys.
    """
    keys = dict_list[0].keys()
    assert all(
        d.keys() == keys for d in dict_list
    ), "All dictionaries must have the same keys."

    return {
        key: torch.mean(torch.stack(values))
        for key, values in zip(dict_list[0], zip(*[d.values() for d in dict_list]))
    }


def seqconfmx_to_macro_ave_sns(seqconfmx):
    """Calc confusion-matrix-based performance metrics.
        V2 supports accuracy and implements a different calc criterion
        of the macro-averaged metrics. V2's output is np.ndarray, not Tensor.
    Important Note (February 24th, 2021):
        If no example belonging to a class comes in `seqconfmx`,
        i.e., all the entries in the raw corresponding to that class are zero,
        then the classwise metric of that class is assumed to be ZERO.
        However, the macro-averaged metric IGNORES such empty classes (V2),
        while in `seqconfmx_to_metrics` (V1), the macro-averaged metrics
        assume that the classwise metrics of the empty classes are ZERO
        (i.e., do not ignore them), which may significantly degradate
        the macro-averaged metrics (e.g., when the sample size used for `seqconfmx`
        is much smaller than the number of classes).
    Args:
        seqconfmx: A series of confusion matrix Tensors
            with shape (series length (arbitrary), num classes, num classes).
    Return:
        dict_metrics: A dictionary of performance metrics including
            classwise, macro-averaged, and micro-averaged metrics.
    Remarks:
        - Examples:
            dics_metrics["SNS"][k] = sensitivity of class 3, where
                k = 0, 1, 2, ..., num classes - 1.
            dict_metrics["SNS"][num classes] = macro-averaged sensitivity.
            dict_metrics["SNS"][num classes + 1] = micro-averaged sensitivity, which
                is equal to accuracy.
    """
    time_steps = seqconfmx.shape[0]
    num_classes = seqconfmx.shape[-1]
    assert num_classes == seqconfmx.shape[-2]

    seqconfmx = seqconfmx.to(torch.float32)  # avoids overflow
    classwise_sample_size = torch.sum(seqconfmx, dim=2)
    # shape = (time_steps, num cls)

    effective_entries = torch.not_equal(classwise_sample_size, 0)
    mask = torch.where(effective_entries)
    # tuple of (column#0 indices, column#1 indices)
    mask = torch.stack(mask, dim=1)
    # A Tensor of integer indices.
    # shape = (num of classes with non-zero sample sizes, 2)
    # 2 means [row index, column index]
    # E.g.,
    #    <tf.Tensor: id=212, shape=(1250, 2), dtype=int64, numpy=
    #    array([[  0,   1],
    #        [  0,   5],
    #        [  0,   6],
    #        ...,
    #        [ 49,  90],
    #        [ 49,  99],
    #        [ 49, 100]])>
    # Usage:
    #    NEW_TENSOR = tf.reshape(tf.gather_nd(TENSOR, mask), [time_steps, -1]),
    #    where TENSOR.shape is (time_steps, num_classes).
    #    NEW_TENSOR has shape (time_steps, num of classes with non-zero sample sizes).
    #    TENSOR can be SNS, TP, TN, FP, FN, etc., as shown below.

    # Calc 2x2 confusion matrices out of the multiclass confusion matrix

    TP = torch.diagonal(seqconfmx, dim1=1, dim2=2)
    # (time_steps, num cls)
    FP = torch.sum(seqconfmx, dim=1) - TP
    # (time_steps, num cls)
    FN = torch.sum(seqconfmx, dim=2) - TP
    # (time_steps, num cls)
    TN = torch.sum(seqconfmx, dim=1)
    # (time_steps, num cls)
    TN = torch.sum(TN, dim=1, keepdims=True)
    # (time_steps, 1)
    TN = TN.repeat(1, num_classes) - (TP + FP + FN)
    # (time_steps, num cls)

    # Sensitivity (Recall, Classwise accuracy)
    ############################################################
    # Calc classwise, macro-ave, micro-ave metrics
    SNS = TP / (TP + FN + 1e-12)
    # (time_steps, num cls)
    gathered_SNS = SNS[mask[:, 0], mask[:, 1]]

    reshaped_SNS = gathered_SNS.reshape(time_steps, -1)

    macroave_sns = torch.mean(reshaped_SNS, dim=1, keepdims=True)
    # (time_steps, 1)

    return macroave_sns


################################################
# Functions For SPRT
################################################


# Functions: Calc LLRs (logit -> LLR)
def calc_llrs(logits_concat):
    """Calculate the frame-by-frame log-likelihood ratio matrices.
        Used to calculate LLR(x^(1), ..., x^(t))
        with N-th order posteriors (the N-th order TANDEM formula).
    Args:
        logits_concat: A logit Tensor with shape
            (batch, (time_steps - order_sprt), order_sprt + 1, num classes).
            This is the output from
            utils.data_processing.sequential_concat(
            logit_slice, labels_slice).
    Returns:
        llr_matrix: A Tensor
            with shape (batch, time_steps, num classes, num classes).
            matrix[i, j] = log(p_i/p_j), where p_i and p_j are the likelihood
            of class i and that of j, resp.
            They are anti-symmetric matrices.
    Remarks:
        - The LLRs returned are the LLRs used in the
          "order_sprt"-th order SPRT; the LLRs unnecessary to calculate the
          "order_sprt"-th order SPRT are not included.
        - "time_steps" and "order_sprt" are automatically calculated
          using logits_concat.shape.
    """
    logits_concat_shape = logits_concat.shape
    order_sprt = int(logits_concat_shape[2] - 1)
    time_steps = int(logits_concat_shape[1] + order_sprt)
    num_classes = int(logits_concat_shape[3])
    assert num_classes > 1, "num_classes={} must > 1".format(num_classes)
    logits1 = torch.unsqueeze(logits_concat, dim=-1)
    # (batch, time_steps - order, order + 1, num cls, 1)
    logits2 = torch.unsqueeze(logits_concat, dim=-2)
    # (batch, time_steps - order, order + 1, 1, num cls)
    list_llrs = []

    # i.i.d. LLR (for 0th order SPRT)
    if order_sprt == 0:
        llrs_all_frames = (
            logits1[:, :, order_sprt, :, 0:] - logits2[:, :, order_sprt, 0:, :]
        )
        # (batch, time_steps, num cls, num cls)
        for iter_frame in range(time_steps):
            llrs = torch.sum(llrs_all_frames[:, : iter_frame + 1, :, :], dim=1)
            # (batch, num cls, num cls)
            list_llrs.append(torch.unsqueeze(llrs, dim=1))

    # N-th order LLR (for N-th order SPRT)
    else:
        for iter_frame in range(time_steps):
            if iter_frame < order_sprt + 1:
                llrs = (
                    logits1[:, 0, iter_frame, :, 0:] - logits2[:, 0, iter_frame, 0:, :]
                )
                # (batch, num cls, num cls)
                list_llrs.append(torch.unsqueeze(llrs, dim=1))

            else:
                llrs1 = (
                    logits1[:, : iter_frame - order_sprt + 1, order_sprt, :, 0:]
                    - logits2[:, : iter_frame - order_sprt + 1, order_sprt, 0:, :]
                )
                # (batch, iter_frame-order_sprt, num cls, num cls)
                llrs1 = torch.sum(llrs1, dim=1)  # (batch, num cls, num cls)
                llrs2 = (
                    logits1[:, 1 : iter_frame - order_sprt + 1, order_sprt - 1, :, 0:]
                    - logits2[:, 1 : iter_frame - order_sprt + 1, order_sprt - 1, 0:, :]
                )
                # (batch, iter_frame-order_sprt-1, num cls, num cls)
                llrs2 = torch.sum(llrs2, dim=1)  # (batch, num cls, num cls)
                llrs = llrs1 - llrs2  # (batch, num cls, num cls)
                list_llrs.append(torch.unsqueeze(llrs, dim=1))

    # (batch, time_steps, num cls, num cls)
    llr_matrix = torch.cat(list_llrs, dim=1)

    return llr_matrix


def calc_llr_abserr(estimated_llrs, ground_truth_llrs):
    """
    Calculate absolute estimation error of LLR matrix.
    The estimated and ground-truth LLR matrix must have the same shape.

    Args:
    - estimated_llrs (numpy array): an estimated likelihood ratio matrix
      of size (batch_size, time_steps, num_classes, num_classes)
    - ground_truth_llrs (numpy array): the ground-truth likelihood ratio matrix
      of size (batch_size, time_steps, num_classes, num_classes)

    Returns:
    - abserr (numpy array): absolute error matrix of the same size as the LLR matrices.
    """
    assert estimated_llrs.shape == ground_truth_llrs.shape

    abserr = torch.abs(estimated_llrs - ground_truth_llrs)
    return abserr


def calc_urgency_signal(u_sgnl_concat):
    """Calculate the urgency signal from fragmented signals.
    Args:
    - u_sgnl_concat: A Tensor with shape (bs, time-order_sprt, order_sprt+1, 1)
    Returns:
    - u_sgnl: A Tensor with shape (bs, time, 1)
    Raises:
    """
    _device = u_sgnl_concat.device
    sgnl_shape = u_sgnl_concat.shape
    bs = sgnl_shape[0]
    order_sprt = sgnl_shape[2] - 1
    time = sgnl_shape[1] + order_sprt

    u_sgnl = torch.zeros((bs, time, 1), device=_device)
    for i in range(sgnl_shape[1]):
        u_sgnl[:, i : i + order_sprt + 1, :] += u_sgnl_concat[:, i, :, :]

    for i in range(time):
        if i < order_sprt and (time - i) > order_sprt:  # at the beginning
            u_sgnl[:, i, :] /= i + 1
        elif i >= order_sprt and (time - i) > order_sprt:  # at the middle
            u_sgnl[:, i, :] /= order_sprt + 1
        elif (time - i) <= order_sprt:  # at the end
            u_sgnl[:, i, :] /= time - i

    return u_sgnl


def calc_oblivious_llrs(logits_concat):
    """Calculate the frame-by-frame log-likelihood ratio matrices.
        Used to calculate LLR(x^(t-N), ..., x^(t))
        i.e., (the N-th order TANDEMsO formula).
    Args:
        logits_concat: A logit Tensor with shape
            (batch, (time_steps - order_sprt), order_sprt + 1, num classes).
            This is the output from
            utils.data_processing.sequential_concat(
            logit_slice, labels_slice).
    Returns:
        llr_matrix: A Tensor
            with shape (batch, time_steps, num classes, num classes).
            matrix[i, j] = log(p_i/p_j), where p_i and p_j are the likelihood
            of class i and that of j, resp.
            They are anti-symmetric matrices.
    Remarks:
        - The LLRs returned are the LLRs used in the
          "order_sprt"-th order SPRT; the LLRs unnecessary to calculate the
          "order_sprt"-th order SPRT are not included.
        - "time_steps" and "order_sprt" are automatically calculated
          from logits_concat.shape.
    """
    logits_concat_shape = logits_concat.shape
    order_sprt = int(logits_concat_shape[2] - 1)
    time_steps = int(logits_concat_shape[1] + order_sprt)
    num_classes = int(logits_concat_shape[3])
    assert num_classes > 1, "num_classes={} must > 1".format(num_classes)

    logits1 = torch.unsqueeze(logits_concat, dim=-1)
    # (batch, time_steps - order, order + 1, num cls, 1)
    logits2 = torch.unsqueeze(logits_concat, dim=-2)
    # (batch, time_steps - order, order + 1, 1, num cls)
    list_llrs = []

    # i.i.d. SPRT (0th-order SPRT)
    if order_sprt == 0:
        llrs_all_frames = (
            logits1[:, :, order_sprt, :, 0:] - logits2[:, :, order_sprt, 0:, :]
        )
        # (batch, time_steps, num cls, num cls)
        llr_matrix = llrs_all_frames  # oblivious!!

    # N-th order LLR (for N-th order oblivious SPRT)
    else:
        for iter_frame in range(time_steps):
            if iter_frame < order_sprt + 1:
                llrs = (
                    logits1[:, 0, iter_frame, :, 0:] - logits2[:, 0, iter_frame, 0:, :]
                )
                # (batch, num cls, num cls)
                list_llrs.append(torch.unsqueeze(llrs, 1))

            else:
                llrs1 = (
                    logits1[:, iter_frame - order_sprt, order_sprt, :, 0:]
                    - logits2[:, iter_frame - order_sprt, order_sprt, 0:, :]
                )
                # (batch, num cls, num cls)
                # removed two colons and two "+1" to be oblivious
                # llrs1 = torch.sum(llrs1, 1) # (batch, num cls, num cls)
                # llrs2 = logits1[:, 1:iter_frame - order_sprt + 1, order_sprt-1, :, 0:]\
                #    - logits2[:, 1:iter_frame - order_sprt + 1, order_sprt-1, 0:, :]
                #    # (batch, iter_frame-order_sprt-1, num cls, num cls)
                # llrs2 = torch.sum(llrs2, 1) # (batch, num cls, num cls)
                llrs = llrs1  # - llrs2 # (batch, num cls, num cls)
                list_llrs.append(torch.unsqueeze(llrs, 1))

        # (batch, time_steps, num cls, num cls)
        llr_matrix = torch.cat(list_llrs, dim=1)

    return llr_matrix


# Functions: threshold (LLR -> thresh)
def threshold_generator(llrs, num_thresh, sparsity):
    """Generates sequences of sigle-valued threshold matrices.
    Args:
        llrs: A Tensor with shape
            [batch, time_steps, num classes, num classes].
            Anti-symmetric matrix.
        num_thresh: An integer, the number of threholds.
            1 => thresh = minLLR
            2 => thresh = minLLR, maxLLR
            3 => thresh = minLLR, (minLLR+maxLLR)/2, maxLLR
            ... (linspace float numbers).
        sparsity: "linspace", "logspace", "unirandom", or "lograndom".
            Linearly spaced, logarithmically spaced, uniformly random,
            or log-uniformly random thresholds are generated
            between min LLR and max LLR.
    Returns:
        thresh: A Tensor with shape
            (num_thresh, time_steps, num classes, num classes).
            In each matrix,
            diag = 0, and off-diag shares a single value > 0.
            Sorted in ascending order of the values.
    Remarks:
        - The threshold values are in [min |LLR| (!= 0), max |LLR|].
        - For reference, we show the Wald's approximation:
          If alpha is a float in (0, 0.5) (FPR) and
          beta is a float in (0, 0.5) (FNR),
          then thresh
          = [np.log(beta/(1-alpha)), np.log((1-beta)/alpha)].
    """
    _device = llrs.device

    llrs_shape = llrs.shape
    num_classes = llrs_shape[-1]

    # Remove 0 LLRs in off-diags
    tri = torch.ones_like(llrs)
    triu = torch.triu(tri)  # Upper triangular part.
    tril = torch.tril(tri)  # Lower triangular part.
    llrs -= 1e-12 * (triu - tril)

    # Calc non-zero max and min of |LLRs|
    llrs_abs = torch.abs(llrs)
    llrs_max = torch.max(llrs_abs)
    # max |LLRs|
    tmp = torch.eye(num_classes, device=_device) * llrs_max
    tmp = tmp.reshape([1, 1, num_classes, num_classes])
    llrs_min = torch.min(llrs_abs + tmp)
    # strictly positive (non-zero) minimum of LLRs

    assert llrs_min > 0.0

    # Single-valued threshold matrix
    if sparsity == "linspace":
        thresh = torch.linspace(
            llrs_min.item(), llrs_max.item(), num_thresh, device=_device
        )
        # (num thresh,)

    elif sparsity == "logspace":
        thresh = torch.exp(
            torch.linspace(
                torch.log(llrs_min).item(),
                torch.log(llrs_max).item(),
                num_thresh,
                device=_device,
            )
        )
        # (num thresh,)

    elif sparsity == "unirandom":
        thresh = torch.rand(num_thresh, device=_device)
        thresh = torch.sort(thresh)[0]
        thresh = ((llrs_max - llrs_min) * thresh) + llrs_min
        # (num thresh,), ascending order

    elif sparsity == "lograndom":
        thresh = torch.rand(num_thresh, device=_device)
        thresh = torch.sort(thresh)[0]
        thresh = torch.exp(
            ((torch.log(llrs_max) - torch.log(llrs_min)) * thresh) + torch.log(llrs_min)
        )
        # (num thresh,). Ascending order.
    else:
        raise ValueError

    return thresh


def tile_constant_threshold(
    thresh: torch.Tensor, batch_size: int, time_steps: int, num_classes: int
) -> torch.Tensor:
    """ """
    _device = thresh.device
    num_thresh = thresh.shape[0]

    # convert to (num thresh, batch_size, time_steps, num cls, num cls)
    thresh = thresh.view(num_thresh, 1, 1, 1, 1)
    thresh = thresh.tile(1, batch_size, time_steps, num_classes, num_classes).to(
        _device
    )

    # diagonal all-zero, else one
    mask = 1 - torch.eye(num_classes, device=_device)

    thresh = mask * (thresh + 1e-11)
    # Avoids 0 threholds, which may occur
    # when logits for different classes have the same value,
    # e.g., 0, due to loss of significance.
    # This operation may cause sparsity of SAT curve
    # if llr_min is << 1e-11, but such a case is ignorable
    # in practice, according to my own experience.
    return thresh  # (num_thresh, time_steps, num classes, num classes).


def thresh_truncated_MSPRT(thresh: torch.Tensor) -> torch.Tensor:
    """
    Set the last threshold in the time dimension to zero, with a small value on off-diagonal elements.

    Args:
        thresh: A PyTorch Tensor with shape
            (num thresholds, time_steps, num class, num class) or
            (batch_size, num thresholds, time_steps, num class, num class).

    Returns:
        A PyTorch Tensor with the same shape as the input, with the last threshold in the
        time dimension set to zero (with a small value on off-diagonal elements).
    """
    ndim = len(thresh.shape)
    if ndim not in [4, 5]:
        raise ValueError("thresh must have 4 or 5 dimensions")

    _device = thresh.device
    num_classes = thresh.shape[-1]

    # diagonal all-zero, else one
    mask = 1 - torch.eye(num_classes, device=_device)

    last_thresh_shape = list(thresh.shape[:-3]) + [1, num_classes, num_classes]
    last_thresh = torch.zeros(last_thresh_shape, device=_device)
    last_thresh += mask * 1e-11

    return torch.cat([thresh[..., :-1, :, :], last_thresh], dim=-3)


def get_upper_triangle(scores_full):
    """
    Get the upper triangle of the scores matrix.

    Args:
        scores_full: A PyTorch Tensor with shape
            (batch_size, num thresholds, time_steps, num_classes, num_classes).
    Returns:
        scores_vec: A PyTorch Tensor with shape
            (batch_size, num thresholds, time_steps, kC2) where k is the num_classes.
    """
    assert len(scores_full.shape) == 5, "scores_full must have 5 dimensions"
    num_classes = scores_full.shape[-1]

    upper_triangle_indices = torch.triu_indices(num_classes, num_classes, offset=1)

    scores_vec = scores_full[
        :, :, :, upper_triangle_indices[0], upper_triangle_indices[1]
    ]

    return scores_vec


def hyperspherical_threshold(
    thresh: torch.Tensor,
    kappa: torch.Tensor,
) -> torch.Tensor:
    """
    Create a tapering threshold with hypersphere function.

    Args:
    - thresh (PyTorch Tensor): (num_thresh, batch_size, time_steps, num_classes, num_classes)
    - kappa (PyTorch Tensor): (num_thresh, batch_size, time_steps, 1)

    Returns:
    - tapering_thresh (PyTorch Tensor): (num_thresh, batch_size, time_steps, num_classes, num_classes)
    """
    assert len(kappa.shape) == 4 and kappa.shape[-1] == 1
    assert len(thresh.shape) == 5

    num_thresh, batch_size, time_steps, _, num_classes = thresh.shape
    _device = thresh.device

    # time steps
    t = torch.linspace(0, time_steps - 1, time_steps, device=_device)
    # expand to 5-dim
    t = t.view(1, 1, -1, 1, 1)

    # expand to 5-dim
    kappa = kappa.view(num_thresh, batch_size, time_steps, 1, 1)

    # Broadcasting and element-wise operations
    tapering_thresh = thresh * (1 - t / (time_steps - 1)) ** (torch.exp(kappa))

    # set diagonal to zero
    mask = 1 - torch.eye(num_classes, device=_device)
    tapering_thresh = mask * (tapering_thresh + 1e-11)

    return tapering_thresh


def check_diag_zeros_positive_offdiag(thresh_mtx: torch.Tensor) -> None:
    """
    Check if the diagonal elements of a threshold matrix are zero and
    the off-diagonal elements are positive.

    Args:
        thresh_mtx: A PyTorch Tensor with shape
            (num thresholds, time_steps, num class, num class) or
            (batch_size, num thresholds, time_steps, num class, num class).

    Raises:
        ValueError: If the diagonal elements of thresh_mtx are not 0.
        ValueError: If the off-diagonal elements of thresh_mtx are not positive.
    """
    num_classes = thresh_mtx.shape[-1]
    _device = thresh_mtx.device
    ndim = len(thresh_mtx.shape)

    if ndim not in [4, 5]:
        raise ValueError("thresh_mtx must have 4 or 5 dimensions")

    for i in range(num_classes):
        diag_mask = [slice(None)] * (ndim - 2) + [i, i]
        if not (thresh_mtx[tuple(diag_mask)] == 0).all():
            raise ValueError(
                "The diag elements of thresh_mtx must be 0."
                + f"\nNow thresh_mtx = {thresh_mtx}"
            )

    tmp = torch.eye(num_classes, device=_device)
    tmp = tmp.reshape(1, 1, *([1] * (ndim - 4)), num_classes, num_classes)
    tmp_th = thresh_mtx + tmp.expand_as(thresh_mtx)
    if not (tmp_th > 0).all():
        raise ValueError(
            "The off-diag elements of thresh_mtx must be positive."
            + f"\nNow thresh_mtx = {thresh_mtx}"
        )


# Function: Matrix SPRT (LLR, thresh -> confmx)
def truncated_MSPRT(llr_mtx, labels_concat, thresh_mtx):
    """Truncated Matrix-SPRT.
    Args:
        llr_mtx: A Tensor with shape
            (batch, time_steps, num classes, num classes).
            Anti-symmetric matrices.
        labels_concat: A Tensor with shape (batch,).
        thresh_mtx: A Tensor with shape
            (num thresholds, time_steps, num class, num class).
            Diag must be 0. Off diag must be strictly positive.
            To be checked in this function.
    Returns:
        confmx: A Tensor with shape (num thresh, classes, num classes).
            Confusion matrix.
        mht: A Tensor with shape (num thresh,). Mean hitting time as a whole.
        vht: A Tensor with shape (num thresh,). Variance of hitting times as a whole.
        trt: A Tensor with shape (num thresh,). Truncation rete as a whole.
    """
    """ check shape match, then values """
    """ care about exactly zero LLRs: Done """
    thresh_mtx_shape = thresh_mtx.shape
    num_thresh = thresh_mtx_shape[0]
    time_steps = thresh_mtx_shape[1]
    num_classes = thresh_mtx_shape[2]
    batch_size = llr_mtx.shape[0]

    # Sanity check of thresholds
    check_diag_zeros_positive_offdiag(thresh_mtx)
    thresh_mtx = thresh_mtx.to(torch.float32)

    # Reshape and calc scores
    llr_mtx = torch.unsqueeze(llr_mtx, dim=0)
    # (1, batch, time_steps, num cls, num cls)
    # to admit the num-thresh axis.
    thresh_mtx = torch.unsqueeze(thresh_mtx, dim=1)
    # (num thresh, 1, time_steps, num cls, num cls)
    # to admit the batch axis.
    tri = torch.ones_like(llr_mtx)
    triu = torch.triu(tri)  # Upper triangular part.
    tril = torch.tril(tri)  # Lower triangular part.
    llr_mtx -= 1e-10 * (triu - tril)
    # (1, batch, time_steps , num cls, num cls)
    # To avoid double hit due to the values exactly equal to zero
    # in scores or when doing truncation, LLRs of the last frame.
    scores = torch.min(llr_mtx - thresh_mtx, -1)
    # (num thresh, batch, time_steps, num cls)
    # Values are non-positive.
    """ assert 1: for each thresh, batch, and time_steps, 
                  the num of 0 is 0 or at most 1 in the last axis direction
        assert 2: values <= 0
    """

    # Calc all predictions and waits
    preds_all = torch.sign(scores) + 1
    # 0:wait, 1:hit (one-hot vector)
    # (num thresh, batch, time_steps, num cls)
    """assert actually one-hot"""

    # Calc truncation rate
    hit_or_wait_all_frames = 1 - preds_all  # wait=1, hit=0
    trt = torch.mean(torch.prod(hit_or_wait_all_frames, dim=(2, 3)), dim=1)
    # (num thresh,)

    if time_steps == 1:
        # Forced decision
        preds_last = torch.sign(torch.min(llr_mtx, -1)) + 1
        # 0: wait, 1: hit (one-hot vector)
        # (1, batch, time_steps=1, num cls)
        """assert check shape"""
        """assert check all the data points in the batch is are now one-hot vectors."""
        preds_last = preds_last.repeat(num_thresh, 1, 1, 1)
        preds_all_trunc = preds_last
        # (num thresh, batch, 1, num cls)

        # Calc hitting times
        mht = torch.tensor(1.0, dtype=torch.float32)
        vht = torch.tensor(0.0, dtype=torch.float32)

        # Calc confusion matrices
        preds = preds_all_trunc[:, :, 0, :]
        # (num thresh, batch, 1, num cls): one-hot vectors

        labels_oh = torch.nn.functional.one_hot(labels_concat, num_classes)  # dim=1
        # (batch, num cls)
        labels_oh = torch.unsqueeze(labels_oh, dim=0)
        labels_oh = labels_oh.repeat(num_thresh, 1, 1)
        # (num thresh, batch, num cls)

        preds = torch.unsqueeze(preds, dim=-2)
        labels_oh = torch.unsqueeze(labels_oh, dim=-1)
        confmx = torch.sum(labels_oh * preds, dim=1).to(torch.int32)
        # (num thresh, num cls, num cls): label axis x pred axis

    else:
        # Forced decision
        preds_last = torch.sign(torch.min(llr_mtx[:, :, -1, :, :], -1)) + 1
        # 0: wait, 1: hit (one-hot vector)
        # (1, batch, num cls)
        """assert check shape"""
        """assert check all the data points in the batch is are now one-hot vectors."""
        preds_last = torch.unsqueeze(preds_last, 2)
        # (1, batch, 1, num cls)
        preds_last = preds_last.repeat(num_thresh, 1, 1, 1)
        # (num thresh, batch, 1, num cls)
        preds_all_trunc = torch.cat([preds_all[:, :, :-1, :], preds_last], dim=2)
        # (num thresh, batch, time_steps - 1, num cls)
        # + (num thresh, batch, 1, num cls)
        # = (num thresh, batch, time_steps, num cls)
        # Now, preds_all_trunc[i, j, t, :] for fixed i and j is
        # a one-hot vector for t = time_steps - 1
        # and
        # filled with 0 or a one-hot vector for t != time_steps - 1.
        """ assert: check this """

        # Calc mean hitting time
        mask = torch.tensor(
            [i + 1 for i in range(time_steps)][::-1], dtype=torch.float32
        )
        mask = mask.repeat(num_thresh * batch_size * num_classes)
        mask = mask.reshape(num_thresh, batch_size, num_classes, time_steps)
        mask = mask.permute(0, 1, 3, 2)
        masked = preds_all_trunc * mask
        # (num thresh, batch, time_steps, num cls)
        hitidx = torch.max(masked, dim=2)
        # (num thresh, batch, num cls)
        hittimes = time_steps - torch.max(hitidx, dim=2) + 1
        # (num thresh, batch)
        mht = torch.mean(hittimes, dim=1)
        vht = torch.var(hittimes, dim=1)
        # (num thresh,)

        # Calc confusion matrix
        preds = torch.argmax(hitidx, dim=2)
        # (num thresh, batch,)
        preds = torch.nn.functional.one_hot(preds, num_classes)  # dim=2
        # (num thresh, batch, num cls)

        labels_oh = torch.nn.functional.one_hot(labels_concat, num_classes)  # dim=1
        # (batch, num cls)
        labels_oh = torch.unsqueeze(labels_oh, dim=0)
        labels_oh = labels_oh.repeat(num_thresh, 1, 1)
        # (num thresh, batch, num cls)

        preds = torch.unsqueeze(preds, dim=-2)
        labels_oh = torch.unsqueeze(labels_oh, dim=-1)
        confmx = torch.sum(labels_oh * preds, dim=1).to(torch.int32)
        # (num thresh, num cls, num cls): label axis x pred axis

    return confmx, mht, vht, trt


# Functions: LLR -> confmx
def llr_sequential_confmx(llrs, labels_concat):
    """For optuna and NP test.
        Calculate the frame-by-frame confusion matrices
        based on the log-likelihood ratios.
    Args:
        llrs: A Tensor
            with shape (batch, time_steps, num classes, num classes).
        labels_concat: A non-one-hot label Tensor with shape (batch,).
            This is the output from
            utils.data_processing.sequential_concat(logit_slice, labels_slice).
    Returns:
        seqconfmx_llr: A Tensor with shape (time_steps, num classes, num classes).
            The sequential confusion matrices of framewise LLRs with thresh=0.
    """
    llrs_shape = llrs.shape
    time_steps = llrs_shape[1]
    num_classes = llrs_shape[2]

    # To avoid double hit due to the values exactly equal to zero in LLRs.
    tri = torch.ones_like(llrs)
    triu = torch.triu(tri)  # Upper triangular part.
    tril = torch.tril(tri)  # Lower triangular part.
    llrs -= 1e-12 * (triu - tril)
    # (batch, time_steps , num cls, num cls)

    # Forced decision
    preds = torch.sign(torch.min(llrs, -1)[0]) + 1
    # 0: wait, 1: hit (one-hot vector)
    # (batch, time_steps, num cls)

    # Calc confusion matrices
    labels_oh = torch.nn.functional.one_hot(labels_concat, num_classes)  # dim=1
    # (batch, num cls)
    labels_oh = torch.unsqueeze(labels_oh, dim=1)
    labels_oh = labels_oh.repeat(1, time_steps, 1)
    # (batch, time_steps, num cls)

    preds = torch.unsqueeze(preds, dim=-2)
    labels_oh = torch.unsqueeze(labels_oh, dim=-1)
    seqconfmx = torch.sum(labels_oh * preds, dim=0).to(torch.int32)
    # (time_steps, num cls, num cls): label axis x pred axis

    return seqconfmx


def NP_test(llrs, labels, length=None):
    """Neyman-Pearson Test.
    Args:
        llrs: A Tensor
            with shape (batch, time_steps, num classes, num classes).
        labels: A Tensor with shape (batch,).
        length: A integer or None. If this is not None,
            it shouldb be 1 <= length <= time_steps,
            and scores after the length-th frame are thrown away.
    Returns:
        A Tensor with shape (num classes, num classes) if length is not None,
        else (time_steps, num classes, num classes).
    Remark:
        - Currently, only threshold = 0 is supported.
        - Note that the NP test uses the likelihood ratio of the sequence,
          not that of a frame in the sequence;
          that's why the input arg must be LLRs, not arbitrary scores.
    """
    if length is None:
        return llr_sequential_confmx(llrs, labels)

    else:
        assert 1 <= length <= llrs.shape[1]
        return llr_sequential_confmx(llrs, labels)[length - 1]


# Functions: Utils
def confmx_to_bac(confmx):
    """confx -> BAC
    Args:
        confmx: A Tensor with shape (num classes, num classes)
    Returns:
        bac: A scalar Tensor, balanced accuracy.
            Independend of num of classes (>= 2).
    """
    dict_metrics = confmx_to_metrics(confmx)
    bac = dict_metrics["SNS"]["macro"]
    return bac


def seqconfmx_to_list_metrics(seqconfmx):
    """Pair function A: sequential confmx -> metrics
    Transforms a Tensor of confusion matirces with shape
    (LEN, num classes, num classes) to a list (with length LEN)
    of dictionaries of metrics, where LEN is undetermined."""
    sequence_length = seqconfmx.shape[0]
    list_metrics = [None] * sequence_length
    for iter_idx in range(sequence_length):
        list_metrics[iter_idx] = confmx_to_metrics(seqconfmx[iter_idx])

    return list_metrics


def list_metrics_to_list_bac(list_metrics):
    """Pair function B: metrics -> BACs
        Input: [confmx_to_metrics(...), confmx_to_metrics(...), ...].
    Arg:
        list_metrics: A list of dictionaries.
    Return:
        list_bacs: A list of floats with the same length as list_metric's.
    """
    list_bacs = [None] * len(list_metrics)
    for iter_idx, iter_dict in enumerate(list_metrics):
        bac = iter_dict["SNS"]["macro"]
        list_bacs[iter_idx] = bac

    return list_bacs


# Functions: other statistical tests
def binary_np_test(llrs, labels, length=None, thresh=0.0):
    """Neyman-Pearson Test.
    Args:
        llrs: A Tensor with shape (batch, time_steps).
        labels: A Tensor with shape (batch,).
        length: A integer or None. If this is not None, it shouldb be 1 <= length <= time_steps, and scores after the length-th frame are thrown away.
        thresh: A float.
    Returns:
        confmx: A Tensor with shape (2, 2). Binary classification is assumed.
    Remark:
        - Note that the NP test uses the likelihood ratio of a sequence, not a frame; that's why the input arg is LLRs, not scores, which is not equivalent to LLRs in general.
    """
    llrs_shape = llrs.shape
    time_steps = llrs_shape[1]
    if not (length is None):
        assert 1 <= length <= time_steps

    # Calc predictions
    llr = llrs[:, length - 1] - thresh
    preds = torch.round(torch.sigmoid(llr))

    # Calc confusion matrix
    confmx = confusion_matrix(labels, preds, num_classes=2)

    return confmx


def binary_avescr_test(scores, labels, length=None, thresh=0.0):
    """Score Average Test.
    Args:
        scores: A Tensor with shape (batch, time_steps).
        labels: A Tensor with shape (batch,).
        length: A integer or None. If this is not None,
            it shouldb be 1 <= length <= time_steps,
            and scores after the length-th frame are thrown away.
        thresh: A float.
    Returns:
        confmx: A Tensor with shape (2, 2). Binary classification is assumed.
    Remark:
        - If scores are drawn from non-i.i.d. distribution,
          e.g., LSTM sequential outputs,
          then temporally averaging scores may be detrimental.
    """
    scores_shape = scores.shape
    time_steps = scores_shape[1]
    if not (length is None):
        assert 1 <= length <= time_steps
        scores = scores[:, :length]

    # Calc predictions
    score = torch.mean(scores, dim=1) - thresh
    preds = torch.round(torch.sigmoid(score))

    # Calc confusion matrix
    confmx = confusion_matrix(labels, preds, num_classes=2)

    return confmx
